import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestOneOf(TorchioTestCase):
    """Tests for `OneOf`."""

    def test_wrong_input_type(self):
        with pytest.raises(ValueError):
            tio.OneOf(1)

    def test_negative_probabilities(self):
        transforms = {
            tio.RandomAffine(): -1,
            tio.RandomElasticDeformation(): 1,
        }
        with pytest.raises(ValueError):
            tio.OneOf(transforms)

    def test_zero_probabilities(self):
        with pytest.raises(ValueError):
            transforms = {
                tio.RandomAffine(): 0,
                tio.RandomElasticDeformation(): 0,
            }
            tio.OneOf(transforms)

    def test_not_transform(self):
        with pytest.raises(ValueError):
            tio.OneOf({tio.RandomAffine: 1, tio.RandomElasticDeformation: 2})

    def test_one_of(self):
        transform = tio.OneOf(
            {
                tio.RandomAffine(): 0.2,
                tio.RandomElasticDeformation(max_displacement=0.5): 0.8,
            }
        )
        transform(self.sample_subject)



import pytest
import torch

from torchio import RandomGamma

from ...utils import TorchioTestCase


class TestRandomGamma(TorchioTestCase):
    """Tests for `RandomGamma`."""

    def get_random_tensor_zero_one(self):
        return torch.rand(4, 5, 6, 7)

    def test_with_zero_gamma(self):
        transform = RandomGamma(log_gamma=0)
        tensor = self.get_random_tensor_zero_one()
        transformed = transform(tensor)
        self.assert_tensor_almost_equal(tensor, transformed)

    def test_with_non_zero_gamma(self):
        transform = RandomGamma(log_gamma=(0.1, 0.3))
        tensor = self.get_random_tensor_zero_one()
        transformed = transform(tensor)
        self.assert_tensor_not_equal(tensor, transformed)

    def test_with_high_gamma(self):
        transform = RandomGamma(log_gamma=(100, 100))
        tensor = self.get_random_tensor_zero_one()
        transformed = transform(tensor)
        self.assert_tensor_almost_equal(
            tensor == 1,
            transformed,
        )

    def test_with_low_gamma(self):
        transform = RandomGamma(log_gamma=(-100, -100))
        tensor = self.get_random_tensor_zero_one()
        transformed = transform(tensor)
        self.assert_tensor_almost_equal(
            tensor > 0,
            transformed,
        )

    def test_wrong_gamma_type(self):
        with pytest.raises(ValueError):
            RandomGamma(log_gamma='wrong')



from collections import defaultdict

import numpy as np
import torch

from ....data.subject import Subject
from ....types import TypeRangeFloat
from ....utils import to_tuple
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomGamma(RandomTransform, IntensityTransform):
    r"""Randomly change contrast of an image by raising its values to the power
    :math:`\gamma`.

    Args:
        log_gamma: Tuple :math:`(a, b)` to compute the exponent
            :math:`\gamma = e ^ \beta`,
            where :math:`\beta \sim \mathcal{U}(a, b)`.
            If a single value :math:`d` is provided, then
            :math:`\beta \sim \mathcal{U}(-d, d)`.
            Negative and positive values for this argument perform gamma
            compression and expansion, respectively.
            See the `Gamma correction`_ Wikipedia entry for more information.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    .. _Gamma correction: https://en.wikipedia.org/wiki/Gamma_correction

    .. note:: Fractional exponentiation of negative values is generally not
        well-defined for non-complex numbers.
        If negative values are found in the input image :math:`I`,
        the applied transform is :math:`\text{sign}(I) |I|^\gamma`,
        instead of the usual :math:`I^\gamma`. The
        :class:`~torchio.transforms.RescaleIntensity`
        transform may be used to ensure that all values are positive. This is
        generally not problematic, but it is recommended to visualize results
        on images with negative values. More information can be found on
        `this StackExchange question`_.

        .. _this StackExchange question: https://math.stackexchange.com/questions/317528/how-do-you-compute-negative-numbers-to-fractional-powers

    .. plot::

        import torch
        import torchio as tio
        subject = tio.datasets.FPG()
        subject.remove_image('seg')
        transform = tio.RandomGamma(log_gamma=(-0.3, -0.3))
        transformed = transform(subject)
        subject.add_image(transformed.t1, 'log -0.3')
        transform = tio.RandomGamma(log_gamma=(0.3, 0.3))
        transformed = transform(subject)
        subject.add_image(transformed.t1, 'log 0.3')
        subject.plot()

    Example:
        >>> import torchio as tio
        >>> subject = tio.datasets.FPG()
        >>> transform = tio.RandomGamma(log_gamma=(-0.3, 0.3))  # gamma between 0.74 and 1.34
        >>> transformed = transform(subject)
    """

    def __init__(self, log_gamma: TypeRangeFloat = (-0.3, 0.3), **kwargs):
        super().__init__(**kwargs)
        self.log_gamma_range = self._parse_range(log_gamma, 'log_gamma')

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        arguments: dict[str, dict] = defaultdict(dict)
        for name, image in images_dict.items():
            gammas = [self.get_params(self.log_gamma_range) for _ in image.data]
            arguments['gamma'][name] = gammas
        transform = Gamma(**self.add_base_args(arguments))
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(self, log_gamma_range: tuple[float, float]) -> float:
        gamma = np.exp(self.sample_uniform(*log_gamma_range))
        return gamma


class Gamma(IntensityTransform):
    r"""Change contrast of an image by raising its values to the power
    :math:`\gamma`.

    Args:
        gamma: Exponent to which values in the image will be raised.
            Negative and positive values for this argument perform gamma
            compression and expansion, respectively.
            See the `Gamma correction`_ Wikipedia entry for more information.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    .. _Gamma correction: https://en.wikipedia.org/wiki/Gamma_correction

    .. note:: Fractional exponentiation of negative values is generally not
        well-defined for non-complex numbers.
        If negative values are found in the input image :math:`I`,
        the applied transform is :math:`\text{sign}(I) |I|^\gamma`,
        instead of the usual :math:`I^\gamma`. The
        :class:`~torchio.transforms.preprocessing.intensity.rescale.RescaleIntensity`
        transform may be used to ensure that all values are positive. This is
        generally not problematic, but it is recommended to visualize results
        on image with negative values. More information can be found on
        `this StackExchange question`_.

        .. _this StackExchange question: https://math.stackexchange.com/questions/317528/how-do-you-compute-negative-numbers-to-fractional-powers

    Example:
        >>> import torchio as tio
        >>> subject = tio.datasets.FPG()
        >>> transform = tio.Gamma(0.8)
        >>> transformed = transform(subject)
    """

    def __init__(self, gamma: float, **kwargs):
        super().__init__(**kwargs)
        self.gamma = gamma
        self.args_names = ['gamma']
        self.invert_transform = False

    def apply_transform(self, subject: Subject) -> Subject:
        gamma = self.gamma
        for name, image in self.get_images_dict(subject).items():
            if self.arguments_are_dict():
                assert isinstance(self.gamma, dict)
                gamma = self.gamma[name]
            gammas = to_tuple(gamma, length=len(image.data))
            transformed_tensors = []
            image.set_data(image.data.float())
            for gamma, tensor in zip(gammas, image.data):
                if self.invert_transform:
                    correction = power(tensor, 1 - gamma)
                    transformed_tensor = tensor * correction
                else:
                    transformed_tensor = power(tensor, gamma)
                transformed_tensors.append(transformed_tensor)
            image.set_data(torch.stack(transformed_tensors))
        return subject


def power(tensor, gamma):
    if tensor.min() < 0:
        output = tensor.sign() * tensor.abs() ** gamma
    else:
        output = tensor**gamma
    return output



import copy
import warnings

import torch

import torchio as tio

from ..utils import TorchioTestCase


class TestInvertibility(TorchioTestCase):
    def test_all_random_transforms(self):
        transform = self.get_large_composed_transform()
        # Remove RandomLabelsToImage as it will add a new image to the subject
        for t in transform.transforms:
            if t.name == 'RandomLabelsToImage':
                transform.transforms.remove(t)  # noqa: B038
                break
        # Ignore elastic deformation and gamma warnings during execution
        # Ignore some transforms not invertible
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            transformed = transform(self.sample_subject)
            inverting_transform = transformed.get_inverse_transform()
            transformed_back = inverting_transform(transformed)
        assert transformed.t1.shape == transformed_back.t1.shape
        self.assert_tensor_equal(
            transformed.label.affine,
            transformed_back.label.affine,
        )

    def test_different_interpolation(self):
        def model_probs(subject):
            subject = copy.deepcopy(subject)
            subject.im.set_data(torch.rand_like(subject.im.data))
            return subject

        def model_label(subject):
            subject = model_probs(subject)
            subject.im.set_data(torch.bernoulli(subject.im.data))
            return subject

        transform = tio.RandomAffine(image_interpolation='bspline')
        subject = copy.deepcopy(self.sample_subject)
        tensor = (torch.rand(1, 20, 20, 20) > 0.5).float()  # 0s and 1s
        subject = tio.Subject(im=tio.ScalarImage(tensor=tensor))
        transformed = transform(subject)
        assert transformed.im.data.min() < 0
        assert transformed.im.data.max() > 1

        subject_probs = model_probs(transformed)
        transformed_back = subject_probs.apply_inverse_transform()
        assert transformed_back.im.data.min() < 0
        assert transformed_back.im.data.max() > 1
        transformed_back_linear = subject_probs.apply_inverse_transform(
            image_interpolation='linear',
        )
        assert transformed_back_linear.im.data.min() >= 0
        assert transformed_back_linear.im.data.max() <= 1

        subject_label = model_label(transformed)
        transformed_back = subject_label.apply_inverse_transform()
        assert transformed_back.im.data.min() < 0
        assert transformed_back.im.data.max() > 1
        transformed_back_linear = subject_label.apply_inverse_transform(
            image_interpolation='nearest',
        )
        assert transformed_back_linear.im.data.unique().tolist() == [0, 1]



from .to_orientation import ToOrientation


class ToCanonical(ToOrientation):
    """Reorder the data to be closest to canonical (RAS+) orientation.

    This transform reorders the voxels and modifies the affine matrix so that
    the voxel orientations are nearest to:

    1. First voxel axis goes from left to Right
    2. Second voxel axis goes from posterior to Anterior
    3. Third voxel axis goes from inferior to Superior

    See `NiBabel docs about image orientation`_ for more information.

    Args:
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    .. note:: The reorientation is performed using
        :meth:`~torchio.transforms.preprocessing.to_orientation.ToOrientation`.

    .. _NiBabel docs about image orientation: https://nipy.org/nibabel/image_orientation.html
    """

    def __init__(self, **kwargs):
        super().__init__('RAS', **kwargs)



import copy
import os
import random
import shutil
import tempfile
import unittest
from collections.abc import Sequence
from pathlib import Path
from random import shuffle

import numpy as np
import pytest
import torch

import torchio as tio


class TorchioTestCase(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures, if any."""
        self.dir = Path(tempfile.gettempdir()) / os.urandom(24).hex()
        self.dir.mkdir(exist_ok=True)
        random.seed(42)
        np.random.seed(42)

        registration_matrix = np.array(
            [
                [1, 0, 0, 10],
                [0, 1, 0, 0],
                [0, 0, 1.2, 0],
                [0, 0, 0, 1],
            ]
        )

        subject_a = tio.Subject(
            t1=tio.ScalarImage(self.get_image_path('t1_a')),
        )
        subject_b = tio.Subject(
            t1=tio.ScalarImage(self.get_image_path('t1_b')),
            label=tio.LabelMap(self.get_image_path('label_b', binary=True)),
        )
        subject_c = tio.Subject(
            label=tio.LabelMap(self.get_image_path('label_c', binary=True)),
        )
        subject_d = tio.Subject(
            t1=tio.ScalarImage(
                self.get_image_path('t1_d'),
                pre_affine=registration_matrix,
            ),
            t2=tio.ScalarImage(self.get_image_path('t2_d')),
            label=tio.LabelMap(self.get_image_path('label_d', binary=True)),
        )
        subject_a4 = tio.Subject(
            t1=tio.ScalarImage(self.get_image_path('t1_a'), components=4),
        )
        self.subjects_list = [
            subject_a,
            subject_a4,
            subject_b,
            subject_c,
            subject_d,
        ]
        self.dataset = tio.SubjectsDataset(self.subjects_list)
        self.sample_subject = self.dataset[-1]  # subject_d
        self.subject_4d = self.dataset[1]

    def make_2d(self, subject):
        subject = copy.deepcopy(subject)
        for image in subject.get_images(intensity_only=False):
            image.set_data(image.data[..., :1])
        return subject

    def make_multichannel(self, subject):
        subject = copy.deepcopy(subject)
        for image in subject.get_images(intensity_only=False):
            image.set_data(torch.cat(4 * (image.data,)))
        return subject

    def flip_affine_x(self, subject):
        subject = copy.deepcopy(subject)
        for image in subject.get_images(intensity_only=False):
            image.affine = np.diag((-1, 1, 1, 1)) @ image.affine
        return subject

    def get_inconsistent_shape_subject(self):
        """Return a subject containing images of different shape."""
        subject = tio.Subject(
            t1=tio.ScalarImage(self.get_image_path('t1_inc')),
            t2=tio.ScalarImage(
                self.get_image_path('t2_inc', shape=(10, 20, 31)),
            ),
            label=tio.LabelMap(
                self.get_image_path(
                    'label_inc',
                    shape=(8, 17, 25),
                    binary=True,
                ),
            ),
            label2=tio.LabelMap(
                self.get_image_path(
                    'label2_inc',
                    shape=(18, 17, 25),
                    binary=True,
                ),
            ),
        )
        return subject

    def get_reference_image_and_path(self):
        """Return a reference image and its path."""
        path = self.get_image_path(
            'ref',
            shape=(10, 20, 31),
            spacing=(1, 1, 2),
        )
        image = tio.ScalarImage(path)
        return image, path

    def get_subject_with_partial_volume_label_map(self, components=1):
        """Return a subject with a partial-volume label map."""
        return tio.Subject(
            t1=tio.ScalarImage(
                self.get_image_path('t1_d'),
            ),
            label=tio.LabelMap(
                self.get_image_path(
                    'label_d2',
                    binary=False,
                    components=components,
                ),
            ),
        )

    def get_subject_with_labels(self, labels):
        return tio.Subject(
            label=tio.LabelMap(
                self.get_image_path(
                    'label_multi',
                    labels=labels,
                ),
            ),
        )

    @staticmethod
    def get_unique_labels(data: torch.Tensor) -> set[int]:
        labels = data.unique().tolist()
        return set(labels)

    @staticmethod
    def get_tensor_with_labels(labels: Sequence) -> torch.Tensor:
        tensor = torch.as_tensor(list(labels))
        return tensor.repeat_interleave(2).reshape(1, 1, 1, -1)

    def tearDown(self):
        """Tear down test fixtures, if any."""
        shutil.rmtree(self.dir)

    def get_ixi_tiny(self):
        root_dir = Path(tempfile.gettempdir()) / 'torchio' / 'ixi_tiny'
        return tio.datasets.IXITiny(root_dir, download=True)

    def get_image_path(
        self,
        stem,
        binary=False,
        labels=None,
        shape=(10, 20, 30),
        spacing=(1, 1, 1),
        components=1,
        add_nans=False,
        suffix=None,
        force_binary_foreground=True,
    ):
        shape = (*shape, 1) if len(shape) == 2 else shape
        data = np.random.rand(components, *shape)
        if binary:
            data = (data > 0.5).astype(np.uint8)
            if not data.sum() and force_binary_foreground:
                data[..., 0] = 1
        elif labels is not None:
            data = (data * (len(labels) + 1)).astype(np.uint8)
            new_data = np.zeros_like(data)
            for i, label in enumerate(labels):
                new_data[data == (i + 1)] = label
                if not (new_data == label).sum():
                    new_data[..., i] = label
            data = new_data
        elif self.flip_coin():  # cast some images
            data *= 100
            dtype = np.uint8 if self.flip_coin() else np.uint16
            data = data.astype(dtype)
        if add_nans:
            data[:] = np.nan
        affine = np.diag((*spacing, 1))
        if suffix is None:
            extensions = '.nii.gz', '.nii', '.nrrd', '.img', '.mnc'
            suffix = random.choice(extensions)
        path = self.dir / f'{stem}{suffix}'
        if self.flip_coin():
            path = str(path)
        image = tio.ScalarImage(
            tensor=data,
            affine=affine,
            check_nans=not add_nans,
        )
        image.save(path)
        return path

    def flip_coin(self):
        return np.random.rand() > 0.5

    def get_tests_data_dir(self):
        return Path(__file__).parent / 'image_data'

    def assert_tensor_not_equal(self, *args, **kwargs):  # noqa: N802
        with pytest.raises(AssertionError):
            self.assert_tensor_equal(*args, **kwargs)

    @staticmethod
    def assert_tensor_equal(*args, **kwargs):  # noqa: N802
        torch.testing.assert_close(
            *args,
            rtol=0,
            atol=0,
            check_dtype=False,
            **kwargs,
        )

    @staticmethod
    def assert_tensor_almost_equal(*args, **kwargs):  # noqa: N802
        torch.testing.assert_close(
            *args,
            **kwargs,
            check_dtype=False,
        )

    @staticmethod
    def assert_tensor_all_zeros(tensor):  # noqa: N802
        assert torch.all(tensor == 0)

    def get_large_composed_transform(self):
        all_classes = get_all_random_transforms()
        shuffle(all_classes)
        transforms = [t() for t in all_classes]
        # Hack as default patch size for RandomSwap is 15 and sample_subject
        # is (10, 20, 30)
        for tr in transforms:
            if tr.name == 'RandomSwap':
                tr.patch_size = np.array((10, 10, 10))
        return tio.Compose(transforms)


def get_all_random_transforms():
    transforms_names = [
        name for name in dir(tio.transforms) if name.startswith('Random')
    ]
    classes = [getattr(tio.transforms, name) for name in transforms_names]
    return classes



import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestRandomSampler(TorchioTestCase):
    def test_not_implemented(self):
        sampler = tio.data.sampler.RandomSampler(1)
        with pytest.raises(NotImplementedError):
            sampler(self.sample_subject, 5)
        with pytest.raises(NotImplementedError):
            sampler.get_probability_map(self.sample_subject)



from __future__ import annotations

import torch

from ....data.image import ScalarImage
from ....data.subject import Subject
from ...intensity_transform import IntensityTransform


class Clamp(IntensityTransform):
    """Clamp intensity values into a range :math:`[a, b]`.

    For more information, see :func:`torch.clamp`.

    Args:
        out_min: Minimum value :math:`a` of the output image. If ``None``, the
            minimum of the image is used.
        out_max: Maximum value :math:`b` of the output image. If ``None``, the
            maximum of the image is used.

    Example:
        >>> import torchio as tio
        >>> ct = tio.datasets.Slicer('CTChest').CT_chest
        >>> HOUNSFIELD_AIR, HOUNSFIELD_BONE = -1000, 1000
        >>> clamp = tio.Clamp(out_min=HOUNSFIELD_AIR, out_max=HOUNSFIELD_BONE)
        >>> ct_clamped = clamp(ct)

    .. plot::

        import torchio as tio
        subject = tio.datasets.Slicer('CTChest')
        ct = subject.CT_chest
        HOUNSFIELD_AIR, HOUNSFIELD_BONE = -1000, 1000
        clamp = tio.Clamp(out_min=HOUNSFIELD_AIR, out_max=HOUNSFIELD_BONE)
        ct_clamped = clamp(ct)
        subject.add_image(ct_clamped, 'Clamped')
        subject.plot()
    """

    def __init__(
        self,
        out_min: float | None = None,
        out_max: float | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.out_min, self.out_max = out_min, out_max
        self.args_names = ['out_min', 'out_max']

    def apply_transform(self, subject: Subject) -> Subject:
        for image in self.get_images(subject):
            assert isinstance(image, ScalarImage)
            self.apply_clamp(image)
        return subject

    def apply_clamp(self, image: ScalarImage) -> None:
        image.set_data(self.clamp(image.data))

    def clamp(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.clamp(self.out_min, self.out_max)



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestRandomBiasField(TorchioTestCase):
    def test_no_bias(self):
        transform = tio.RandomBiasField(coefficients=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_almost_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_bias(self):
        transform = tio.RandomBiasField(coefficients=0.1)
        transformed = transform(self.sample_subject)
        self.assert_tensor_not_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_wrong_coefficient_type(self):
        with pytest.raises(ValueError):
            tio.RandomBiasField(coefficients='wrong')

    def test_negative_order(self):
        with pytest.raises(ValueError):
            tio.RandomBiasField(order=-1)

    def test_wrong_order_type(self):
        with pytest.raises(TypeError):
            tio.RandomBiasField(order='wrong')

    def test_small_image(self):
        # https://github.com/TorchIO-project/torchio/issues/300
        tio.RandomBiasField()(torch.rand(1, 2, 3, 4))



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Callable
from typing import Optional
from typing import Union

import numpy as np
import torch

# For typing hints
TypePath = Union[str, Path]
TypeNumber = Union[int, float]
TypeKeys = Optional[Sequence[str]]
TypeData = Union[torch.Tensor, np.ndarray]
TypeDataAffine = tuple[torch.Tensor, np.ndarray]
TypeSlice = Union[int, slice]

TypeDoubletInt = tuple[int, int]
TypeTripletInt = tuple[int, int, int]
TypeQuartetInt = tuple[int, int, int, int]
TypeSextetInt = tuple[int, int, int, int, int, int]

TypeDoubleFloat = tuple[float, float]
TypeTripletFloat = tuple[float, float, float]
TypeSextetFloat = tuple[float, float, float, float, float, float]

TypeTuple = Union[int, TypeTripletInt]
TypeRangeInt = Union[int, TypeDoubletInt]
TypeSpatialShape = Union[int, TypeTripletInt]
TypeRangeFloat = Union[float, tuple[float, float]]
TypeCallable = Callable[[torch.Tensor], torch.Tensor]
TypeDirection2D = tuple[float, float, float, float]
TypeDirection3D = tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
]
TypeDirection = Union[TypeDirection2D, TypeDirection3D]



from .random_bias_field import BiasField
from .random_bias_field import RandomBiasField
from .random_blur import Blur
from .random_blur import RandomBlur
from .random_gamma import Gamma
from .random_gamma import RandomGamma
from .random_ghosting import Ghosting
from .random_ghosting import RandomGhosting
from .random_labels_to_image import LabelsToImage
from .random_labels_to_image import RandomLabelsToImage
from .random_motion import Motion
from .random_motion import RandomMotion
from .random_noise import Noise
from .random_noise import RandomNoise
from .random_spike import RandomSpike
from .random_spike import Spike
from .random_swap import RandomSwap
from .random_swap import Swap

__all__ = [
    'RandomSwap',
    'Swap',
    'RandomBlur',
    'Blur',
    'RandomNoise',
    'Noise',
    'RandomSpike',
    'Spike',
    'RandomGamma',
    'Gamma',
    'RandomMotion',
    'Motion',
    'RandomGhosting',
    'Ghosting',
    'RandomBiasField',
    'BiasField',
    'RandomLabelsToImage',
    'LabelsToImage',
]



from __future__ import annotations

import torch

from ...types import TypeRangeFloat
from ...types import TypeSextetFloat
from ...types import TypeTripletFloat
from ..transform import Transform


class RandomTransform(Transform):
    """Base class for stochastic augmentation transforms.

    Args:
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse_degrees(
        self,
        degrees: TypeRangeFloat,
    ) -> tuple[float, float]:
        return self._parse_range(degrees, 'degrees')

    def parse_translation(
        self,
        translation: TypeRangeFloat,
    ) -> tuple[float, float]:
        return self._parse_range(translation, 'translation')

    @staticmethod
    def sample_uniform(a: float, b: float) -> float:
        return torch.FloatTensor(1).uniform_(a, b).item()

    @staticmethod
    def _get_random_seed() -> int:
        """Generate a random seed.

        Returns:
            A random seed as an int.
        """
        return int(torch.randint(0, 2**31, (1,)).item())

    @staticmethod
    def sample_uniform_sextet(params: TypeSextetFloat) -> TypeTripletFloat:
        results = []
        for a, b in zip(params[::2], params[1::2]):
            results.append(RandomTransform.sample_uniform(a, b))
        sx, sy, sz = results
        return sx, sy, sz



from __future__ import annotations

import copy
import pprint
from collections.abc import Sequence
from typing import TYPE_CHECKING
from typing import Any

import numpy as np

from ..constants import INTENSITY
from ..constants import TYPE
from ..utils import get_subclasses
from .image import Image

if TYPE_CHECKING:
    from ..transforms import Compose
    from ..transforms import Transform


class Subject(dict):
    """Class to store information about the images corresponding to a subject.

    Args:
        *args: If provided, a dictionary of items.
        **kwargs: Items that will be added to the subject sample.

    Example:

        >>> import torchio as tio
        >>> # One way:
        >>> subject = tio.Subject(
        ...     one_image=tio.ScalarImage('path_to_image.nii.gz'),
        ...     a_segmentation=tio.LabelMap('path_to_seg.nii.gz'),
        ...     age=45,
        ...     name='John Doe',
        ...     hospital='Hospital Juan Negrín',
        ... )
        >>> # If you want to create the mapping before, or have spaces in the keys:
        >>> subject_dict = {
        ...     'one image': tio.ScalarImage('path_to_image.nii.gz'),
        ...     'a segmentation': tio.LabelMap('path_to_seg.nii.gz'),
        ...     'age': 45,
        ...     'name': 'John Doe',
        ...     'hospital': 'Hospital Juan Negrín',
        ... }
        >>> subject = tio.Subject(subject_dict)
    """

    def __init__(self, *args, **kwargs: dict[str, Any]):
        if args:
            if len(args) == 1 and isinstance(args[0], dict):
                kwargs.update(args[0])
            else:
                message = 'Only one dictionary as positional argument is allowed'
                raise ValueError(message)
        super().__init__(**kwargs)
        self._parse_images(self.get_images(intensity_only=False))
        self.update_attributes()  # this allows me to do e.g. subject.t1
        self.applied_transforms: list[tuple[str, dict]] = []

    def __repr__(self):
        num_images = len(self.get_images(intensity_only=False))
        string = (
            f'{self.__class__.__name__}'
            f'(Keys: {tuple(self.keys())}; images: {num_images})'
        )
        return string

    def __len__(self):
        return len(self.get_images(intensity_only=False))

    def __getitem__(self, item):
        if isinstance(item, (slice, int, tuple)):
            try:
                self.check_consistent_spatial_shape()
            except RuntimeError as e:
                message = (
                    'To use indexing, all images in the subject must have the'
                    ' same spatial shape'
                )
                raise RuntimeError(message) from e
            copied = copy.deepcopy(self)
            for image_name, image in copied.items():
                copied[image_name] = image[item]
            return copied
        else:
            return super().__getitem__(item)

    @staticmethod
    def _parse_images(images: list[Image]) -> None:
        # Check that it's not empty
        if not images:
            raise TypeError('A subject without images cannot be created')

    @property
    def shape(self):
        """Return shape of first image in subject.

        Consistency of shapes across images in the subject is checked first.

        Example:

            >>> import torchio as tio
            >>> colin = tio.datasets.Colin27()
            >>> colin.shape
            (1, 181, 217, 181)
        """
        self.check_consistent_attribute('shape')
        return self.get_first_image().shape

    @property
    def spatial_shape(self):
        """Return spatial shape of first image in subject.

        Consistency of spatial shapes across images in the subject is checked
        first.

        Example:

            >>> import torchio as tio
            >>> colin = tio.datasets.Colin27()
            >>> colin.spatial_shape
            (181, 217, 181)
        """
        self.check_consistent_spatial_shape()
        return self.get_first_image().spatial_shape

    @property
    def spacing(self):
        """Return spacing of first image in subject.

        Consistency of spacings across images in the subject is checked first.

        Example:

            >>> import torchio as tio
            >>> colin = tio.datasets.Slicer()
            >>> colin.spacing
            (1.0, 1.0, 1.2999954223632812)
        """
        self.check_consistent_attribute('spacing')
        return self.get_first_image().spacing

    @property
    def history(self):
        # Kept for backwards compatibility
        return self.get_applied_transforms()

    def is_2d(self):
        return all(i.is_2d() for i in self.get_images(intensity_only=False))

    def get_applied_transforms(
        self,
        ignore_intensity: bool = False,
        image_interpolation: str | None = None,
    ) -> list[Transform]:
        from ..transforms.intensity_transform import IntensityTransform
        from ..transforms.transform import Transform

        name_to_transform = {cls.__name__: cls for cls in get_subclasses(Transform)}
        transforms_list = []
        for transform_name, arguments in self.applied_transforms:
            transform = name_to_transform[transform_name](**arguments)
            if ignore_intensity and isinstance(transform, IntensityTransform):
                continue
            resamples = hasattr(transform, 'image_interpolation')
            if resamples and image_interpolation is not None:
                parsed = transform.parse_interpolation(image_interpolation)
                transform.image_interpolation = parsed
            transforms_list.append(transform)
        return transforms_list

    def get_composed_history(
        self,
        ignore_intensity: bool = False,
        image_interpolation: str | None = None,
    ) -> Compose:
        from ..transforms.augmentation.composition import Compose

        transforms = self.get_applied_transforms(
            ignore_intensity=ignore_intensity,
            image_interpolation=image_interpolation,
        )
        return Compose(transforms)

    def get_inverse_transform(
        self,
        warn: bool = True,
        ignore_intensity: bool = False,
        image_interpolation: str | None = None,
    ) -> Compose:
        """Get a reversed list of the inverses of the applied transforms.

        Args:
            warn: Issue a warning if some transforms are not invertible.
            ignore_intensity: If ``True``, all instances of
                :class:`~torchio.transforms.intensity_transform.IntensityTransform`
                will be ignored.
            image_interpolation: Modify interpolation for scalar images inside
                transforms that perform resampling.
        """
        history_transform = self.get_composed_history(
            ignore_intensity=ignore_intensity,
            image_interpolation=image_interpolation,
        )
        inverse_transform = history_transform.inverse(warn=warn)
        return inverse_transform

    def apply_inverse_transform(self, **kwargs) -> Subject:
        """Apply the inverse of all applied transforms, in reverse order.

        Args:
            **kwargs: Keyword arguments passed on to
                :meth:`~torchio.data.subject.Subject.get_inverse_transform`.
        """
        inverse_transform = self.get_inverse_transform(**kwargs)
        transformed: Subject
        transformed = inverse_transform(self)  # type: ignore[assignment]
        transformed.clear_history()
        return transformed

    def clear_history(self) -> None:
        self.applied_transforms = []

    def check_consistent_attribute(
        self,
        attribute: str,
        relative_tolerance: float = 1e-6,
        absolute_tolerance: float = 1e-6,
        message: str | None = None,
    ) -> None:
        r"""Check for consistency of an attribute across all images.

        Args:
            attribute: Name of the image attribute to check
            relative_tolerance: Relative tolerance for :func:`numpy.allclose()`
            absolute_tolerance: Absolute tolerance for :func:`numpy.allclose()`

        Example:
            >>> import numpy as np
            >>> import torch
            >>> import torchio as tio
            >>> scalars = torch.randn(1, 512, 512, 100)
            >>> mask = torch.tensor(scalars > 0).type(torch.int16)
            >>> af1 = np.eye([0.8, 0.8, 2.50000000000001, 1])
            >>> af2 = np.eye([0.8, 0.8, 2.49999999999999, 1])  # small difference here (e.g. due to different reader)
            >>> subject = tio.Subject(
            ...   image = tio.ScalarImage(tensor=scalars, affine=af1),
            ...   mask = tio.LabelMap(tensor=mask, affine=af2)
            ... )
            >>> subject.check_consistent_attribute('spacing')  # no error as tolerances are > 0

        .. note:: To check that all values for a specific attribute are close
            between all images in the subject, :func:`numpy.allclose()` is used.
            This function returns ``True`` if
            :math:`|a_i - b_i| \leq t_{abs} + t_{rel} * |b_i|`, where
            :math:`a_i` and :math:`b_i` are the :math:`i`-th element of the same
            attribute of two images being compared,
            :math:`t_{abs}` is the ``absolute_tolerance`` and
            :math:`t_{rel}` is the ``relative_tolerance``.
        """
        message = (
            f'More than one value for "{attribute}" found in subject images:\n{{}}'
        )

        names_images = self.get_images_dict(intensity_only=False).items()
        try:
            first_attribute = None
            first_image = None

            for image_name, image in names_images:
                if first_attribute is None:
                    first_attribute = getattr(image, attribute)
                    first_image = image_name
                    continue
                current_attribute = getattr(image, attribute)
                all_close = np.allclose(
                    current_attribute,
                    first_attribute,
                    rtol=relative_tolerance,
                    atol=absolute_tolerance,
                )
                if not all_close:
                    message = message.format(
                        pprint.pformat(
                            {
                                first_image: first_attribute,
                                image_name: current_attribute,
                            }
                        ),
                    )
                    raise RuntimeError(message)
        except TypeError:
            # fallback for non-numeric values
            values_dict = {}
            for image_name, image in names_images:
                values_dict[image_name] = getattr(image, attribute)
            num_unique_values = len(set(values_dict.values()))
            if num_unique_values > 1:
                message = message.format(pprint.pformat(values_dict))
                raise RuntimeError(message) from None

    def check_consistent_spatial_shape(self) -> None:
        self.check_consistent_attribute('spatial_shape')

    def check_consistent_orientation(self) -> None:
        self.check_consistent_attribute('orientation')

    def check_consistent_affine(self) -> None:
        self.check_consistent_attribute('affine')

    def check_consistent_space(self) -> None:
        try:
            self.check_consistent_attribute('spacing')
            self.check_consistent_attribute('direction')
            self.check_consistent_attribute('origin')
            self.check_consistent_spatial_shape()
        except RuntimeError as e:
            message = (
                'As described above, some images in the subject are not in the'
                ' same space. You probably can use the transforms ToCanonical'
                ' and Resample to fix this, as explained at'
                ' https://github.com/TorchIO-project/torchio/issues/647#issuecomment-913025695'
            )
            raise RuntimeError(message) from e

    def get_images_names(self) -> list[str]:
        return list(self.get_images_dict(intensity_only=False).keys())

    def get_images_dict(
        self,
        intensity_only=True,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
    ) -> dict[str, Image]:
        images = {}
        for image_name, image in self.items():
            if not isinstance(image, Image):
                continue
            if intensity_only and not image[TYPE] == INTENSITY:
                continue
            if include is not None and image_name not in include:
                continue
            if exclude is not None and image_name in exclude:
                continue
            images[image_name] = image
        return images

    def get_images(
        self,
        intensity_only=True,
        include: Sequence[str] | None = None,
        exclude: Sequence[str] | None = None,
    ) -> list[Image]:
        images_dict = self.get_images_dict(
            intensity_only=intensity_only,
            include=include,
            exclude=exclude,
        )
        return list(images_dict.values())

    def get_first_image(self) -> Image:
        return self.get_images(intensity_only=False)[0]

    def add_transform(
        self,
        transform: Transform,
        parameters_dict: dict,
    ) -> None:
        self.applied_transforms.append((transform.name, parameters_dict))

    def load(self) -> None:
        """Load images in subject on RAM."""
        for image in self.get_images(intensity_only=False):
            image.load()

    def unload(self) -> None:
        """Unload images in subject."""
        for image in self.get_images(intensity_only=False):
            image.unload()

    def update_attributes(self) -> None:
        # This allows to get images using attribute notation, e.g. subject.t1
        self.__dict__.update(self)

    @staticmethod
    def _check_image_name(image_name):
        if not isinstance(image_name, str):
            message = (
                f'The image name must be a string, but it has type "{type(image_name)}"'
            )
            raise ValueError(message)
        return image_name

    def add_image(self, image: Image, image_name: str) -> None:
        """Add an image to the subject instance."""
        if not isinstance(image, Image):
            message = (
                'Image must be an instance of torchio.Image,'
                f' but its type is "{type(image)}"'
            )
            raise ValueError(message)
        self._check_image_name(image_name)
        self[image_name] = image
        self.update_attributes()

    def remove_image(self, image_name: str) -> None:
        """Remove an image from the subject instance."""
        self._check_image_name(image_name)
        del self[image_name]
        delattr(self, image_name)

    def plot(self, **kwargs) -> None:
        """Plot images using matplotlib.

        Args:
            **kwargs: Keyword arguments that will be passed on to
                :meth:`~torchio.Image.plot`.
        """
        from ..visualization import plot_subject  # avoid circular import

        plot_subject(self, **kwargs)



import numpy as np
import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestCopyAffine(TorchioTestCase):
    """Tests for `CopyAffine`."""

    def test_missing_reference(self):
        transform = tio.CopyAffine(target='missing')
        with pytest.raises(RuntimeError):
            transform(self.sample_subject)

    def test_wrong_target_type(self):
        with pytest.raises(ValueError):
            tio.CopyAffine(target=[1])

    def test_same_affine(self):
        image = tio.ScalarImage(tensor=torch.rand(2, 2, 2, 2))
        mask = tio.LabelMap(tensor=torch.rand(2, 2, 2, 2))
        mask.affine *= 1.1
        subject = tio.Subject(t1=image, mask=mask)
        transform = tio.CopyAffine('t1')
        transformed = transform(subject)
        self.assert_tensor_equal(
            transformed['t1'].affine,
            transformed['mask'].affine,
        )

    def test_before_loading(self):
        affine = 2 * np.eye(4)
        image = tio.ScalarImage(tensor=torch.rand(2, 2, 2, 2), affine=affine)
        mask = tio.LabelMap(self.get_image_path('mask'))
        subject = tio.Subject(t1=image, mask=mask)
        transform = tio.CopyAffine('t1')
        transformed = transform(subject)
        self.assert_tensor_equal(
            transformed['t1'].affine,
            transformed['mask'].affine,
        )



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestKeepLargestComponent(TorchioTestCase):
    """Tests for `KeepLargestComponent`."""

    def test_one_hot(self):
        tensor = torch.as_tensor([1, 0, 1, 1, 0, 1]).reshape(1, 1, 1, 6)
        label_map = tio.LabelMap(tensor=tensor)
        largest = tio.KeepLargestComponent()(label_map)
        assert largest.data.sum() == 2

    def test_multichannel(self):
        label_map = tio.LabelMap(tensor=torch.rand(2, 3, 3, 3) > 1)
        with pytest.raises(RuntimeError):
            tio.KeepLargestComponent()(label_map)



import copy

import numpy as np
import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestRescaleIntensity(TorchioTestCase):
    def test_rescale_to_same_intentisy(self):
        min_t1 = float(self.sample_subject.t1.data.min())
        max_t1 = float(self.sample_subject.t1.data.max())
        transform = tio.RescaleIntensity(out_min_max=(min_t1, max_t1))
        transformed = transform(self.sample_subject)
        assert np.allclose(
            transformed.t1.data,
            self.sample_subject.t1.data,
            rtol=0,
            atol=1e-05,
        )

    def test_min_max(self):
        transform = tio.RescaleIntensity(out_min_max=(0, 1))
        transformed = transform(self.sample_subject)
        assert transformed.t1.data.min() == 0
        assert transformed.t1.data.max() == 1

    def test_percentiles(self):
        low_quantile = np.percentile(self.sample_subject.t1.data, 5)
        high_quantile = np.percentile(self.sample_subject.t1.data, 95)
        low_indices = (self.sample_subject.t1.data < low_quantile).nonzero(
            as_tuple=True,
        )
        high_indices = (self.sample_subject.t1.data > high_quantile).nonzero(
            as_tuple=True,
        )
        rescale = tio.RescaleIntensity(out_min_max=(0, 1), percentiles=(5, 95))
        transformed = rescale(self.sample_subject)
        assert (transformed.t1.data[low_indices] == 0).all()
        assert (transformed.t1.data[high_indices] == 1).all()

    def test_masking_using_label(self):
        transform = tio.RescaleIntensity(
            out_min_max=(0, 1),
            percentiles=(5, 95),
            masking_method='label',
        )
        transformed = transform(self.sample_subject)
        mask = self.sample_subject.label.data > 0
        low_quantile = np.percentile(self.sample_subject.t1.data[mask], 5)
        high_quantile = np.percentile(self.sample_subject.t1.data[mask], 95)
        low_indices = (self.sample_subject.t1.data < low_quantile).nonzero(
            as_tuple=True,
        )
        high_indices = (self.sample_subject.t1.data > high_quantile).nonzero(
            as_tuple=True,
        )
        assert transformed.t1.data.min() == 0
        assert transformed.t1.data.max() == 1
        assert (transformed.t1.data[low_indices] == 0).all()
        assert (transformed.t1.data[high_indices] == 1).all()

    def test_ct(self):
        ct_max = 1500
        ct_min = -2000
        ct_range = ct_max - ct_min
        tensor = torch.rand(1, 30, 30, 30) * ct_range + ct_min
        ct = tio.ScalarImage(tensor=tensor)
        ct_air = -1000
        ct_bone = 1000
        rescale = tio.RescaleIntensity(
            out_min_max=(-1, 1),
            in_min_max=(ct_air, ct_bone),
        )
        rescaled = rescale(ct)
        assert rescaled.data.min() < -1
        assert rescaled.data.max() > 1

    def test_out_min_higher_than_out_max(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max=(1, 0))

    def test_too_many_values_for_out_min_max(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max=(1, 2, 3))

    def test_wrong_out_min_max_type(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max='wrong')

    def test_min_percentile_higher_than_max_percentile(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max=(0, 1), percentiles=(1, 0))

    def test_too_many_values_for_percentiles(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max=(0, 1), percentiles=(1, 2, 3))

    def test_wrong_percentiles_type(self):
        with pytest.raises(ValueError):
            tio.RescaleIntensity(out_min_max=(0, 1), percentiles='wrong')

    def test_empty_mask(self):
        subject = copy.deepcopy(self.sample_subject)
        subject.label.set_data(subject.label.data * 0)
        rescale = tio.RescaleIntensity(masking_method='label')
        with pytest.warns(RuntimeWarning):
            rescale(subject)

    def test_persistent_in_min_max(self):
        # see https://github.com/TorchIO-project/torchio/issues/1115
        img1 = torch.tensor([[[[0, 1]]]])
        img2 = torch.tensor([[[[0, 10]]]])

        rescale = tio.RescaleIntensity(out_min_max=(0, 1))

        assert rescale(img1).data.flatten().tolist() == [0, 1]
        assert rescale(img2).data.flatten().tolist() == [0, 1]

        rescale = tio.RescaleIntensity(out_min_max=(0, 1))

        assert rescale(img2).data.flatten().tolist() == [0, 1]
        assert rescale(img1).data.flatten().tolist() == [0, 1]



"""
Resample only one axis
======================

In this example, we create a custom preprocessing transfom that changes the
image spacing across one axis only.

Inspired by
`this discussion <https://github.com/TorchIO-project/torchio/discussions/885>`_.
"""

import torch

import torchio as tio


class ResampleZ:
    def __init__(self, spacing_z):
        self.spacing_z = spacing_z

    def __call__(self, subject):
        # We'll assume all images in the subject have the same spacing
        sx, sy, _ = subject.spacing
        resample = tio.Resample((sx, sy, self.spacing_z))
        resampled = resample(subject)
        return resampled


torch.manual_seed(42)
image = tio.datasets.FPG().t1
transforms = tio.ToCanonical(), ResampleZ(spacing_z=7)
transform = tio.Compose(transforms)
transformed = transform(image)
subject = tio.Subject(original=image, transformed=transformed)
subject.plot()



import urllib.parse

from ...data import LabelMap
from ...data import ScalarImage
from ...download import download_and_extract_archive
from ...utils import compress
from .mni import SubjectMNI

SUPPORTED_YEARS = (
    (4.5, 18.5),
    (4.5, 8.5),
    (7, 11),
    (7.5, 13.5),
    (10, 14),
    (13, 18.5),
)


def format_age(n):
    integer = int(n)
    decimal = int(10 * (n - integer))
    return f'{integer:02d}.{decimal}'


class Pediatric(SubjectMNI):
    """MNI pediatric atlases.

    See `the MNI website <https://nist.mni.mcgill.ca/pediatric-atlases-4-5-18-5y/>`_
    for more information.

    .. image:: https://nist.mni.mcgill.ca/wp-content/uploads/2016/04/nihpd_asym_all_sm.jpg
        :alt: Pediatric MNI template

    Arguments:
        years: Tuple of 2 ages. Possible values are: ``(4.5, 18.5)``,
            ``(4.5, 8.5)``,
            ``(7, 11)``,
            ``(7.5, 13.5)``,
            ``(10, 14)`` and
            ``(13, 18.5)``.
        symmetric: If ``True``, the left-right symmetric templates will be
            used. Else, the asymmetric (natural) templates will be used.
    """

    def __init__(self, years, symmetric=False):
        self.url_dir = 'http://www.bic.mni.mcgill.ca/~vfonov/nihpd/obj1/'
        sym_string = 'sym' if symmetric else 'asym'
        if not isinstance(years, tuple) or years not in SUPPORTED_YEARS:
            message = f'Years must be a tuple in {SUPPORTED_YEARS}'
            raise ValueError(message)
        a, b = years
        self.file_id = f'{sym_string}_{format_age(a)}-{format_age(b)}'
        self.name = f'nihpd_{self.file_id}_nifti'
        self.filename = f'{self.name}.zip'
        self.url = urllib.parse.urljoin(self.url_dir, self.filename)
        if not self.download_root.is_dir():
            download_and_extract_archive(
                self.url,
                download_root=self.download_root,
                filename=self.filename,
            )
            (self.download_root / self.filename).unlink()
            for path in self.download_root.glob('*.nii'):
                compress(path)
                path.unlink()

        try:
            subject_dict = self.get_subject_dict('.nii.gz')
        except FileNotFoundError:  # for backward compatibility
            subject_dict = self.get_subject_dict('.nii')
        super().__init__(subject_dict)

    def get_subject_dict(self, extension):
        root = self.download_root
        subject_dict = {
            't1': ScalarImage(root / f'nihpd_{self.file_id}_t1w{extension}'),
            't2': ScalarImage(root / f'nihpd_{self.file_id}_t2w{extension}'),
            'pd': ScalarImage(root / f'nihpd_{self.file_id}_pdw{extension}'),
            'mask': LabelMap(root / f'nihpd_{self.file_id}_mask{extension}'),
        }
        return subject_dict



from ..data import Image
from ..data.subject import Subject
from .transform import Transform


class SpatialTransform(Transform):
    """Transform that modifies image bounds or voxels positions."""

    def get_images(self, subject: Subject) -> list[Image]:
        images = subject.get_images(
            intensity_only=False,
            include=self.include,
            exclude=self.exclude,
        )
        return images



import numpy as np
import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestCropOrPad(TorchioTestCase):
    """Tests for `CropOrPad`."""

    def test_no_changes(self):
        sample_t1 = self.sample_subject['t1']
        shape = sample_t1.spatial_shape
        transform = tio.CropOrPad(shape)
        transformed = transform(self.sample_subject)
        self.assert_tensor_equal(sample_t1.data, transformed['t1'].data)
        self.assert_tensor_equal(sample_t1.affine, transformed['t1'].affine)

    def test_no_changes_mask(self):
        sample_t1 = self.sample_subject['t1']
        sample_mask = self.sample_subject['label'].data
        sample_mask *= 0
        shape = sample_t1.spatial_shape
        transform = tio.CropOrPad(shape, mask_name='label')
        with pytest.warns(RuntimeWarning):
            transformed = transform(self.sample_subject)
        for key in transformed:
            image = self.sample_subject[key]
            self.assert_tensor_equal(image.data, transformed[key].data)
            self.assert_tensor_equal(image.affine, transformed[key].affine)

    def test_different_shape(self):
        shape = self.sample_subject['t1'].spatial_shape
        target_shape = 9, 21, 30
        transform = tio.CropOrPad(target_shape)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            self.assertNotEqual(shape, result_shape)

    def test_shape_right(self):
        target_shape = 9, 21, 30
        transform = tio.CropOrPad(target_shape)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert target_shape == result_shape

    def test_only_pad(self):
        target_shape = 11, 22, 30
        transform = tio.CropOrPad(target_shape)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert target_shape == result_shape

    def test_only_crop(self):
        target_shape = 9, 18, 30
        transform = tio.CropOrPad(target_shape)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert target_shape == result_shape

    def test_shape_negative(self):
        with pytest.raises(ValueError):
            tio.CropOrPad(-1)

    def test_shape_float(self):
        with pytest.raises(ValueError):
            tio.CropOrPad(2.5)

    def test_shape_string(self):
        with pytest.raises(ValueError):
            tio.CropOrPad('')

    def test_shape_one(self):
        transform = tio.CropOrPad(1)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert result_shape == (1, 1, 1)

    def test_wrong_mask_name(self):
        cop = tio.CropOrPad(1, mask_name='wrong')
        with pytest.warns(RuntimeWarning):
            cop(self.sample_subject)

    def test_empty_mask(self):
        target_shape = 8, 22, 30
        transform = tio.CropOrPad(target_shape, mask_name='label')
        mask = self.sample_subject['label'].data
        mask *= 0
        with pytest.warns(RuntimeWarning):
            transform(self.sample_subject)

    def mask_only(self, target_shape):
        transform = tio.CropOrPad(target_shape, mask_name='label')
        mask = self.sample_subject['label'].data
        mask *= 0
        mask[0, 4:6, 5:8, 3:7] = 1
        transformed = transform(self.sample_subject)
        shapes = []
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            shapes.append(result_shape)
        set_shapes = set(shapes)
        message = f'Images have different shapes: {set_shapes}'
        assert len(set_shapes) == 1, message
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert target_shape == result_shape, f'Wrong shape for image: {key}'

    def test_mask_only_pad(self):
        self.mask_only((11, 22, 30))

    def test_mask_only_crop(self):
        self.mask_only((9, 18, 30))

    def test_center_mask(self):
        """The mask bounding box and the input image have the same center."""
        target_shape = 8, 22, 30
        transform_center = tio.CropOrPad(target_shape)
        transform_mask = tio.CropOrPad(target_shape, mask_name='label')
        mask = self.sample_subject['label'].data
        mask *= 0
        mask[0, 4:6, 9:11, 14:16] = 1
        transformed_center = transform_center(self.sample_subject)
        transformed_mask = transform_mask(self.sample_subject)
        zipped = zip(transformed_center.values(), transformed_mask.values())
        for image_center, image_mask in zipped:
            self.assert_tensor_equal(
                image_center.data,
                image_mask.data,
                msg='Data is different after cropping',
            )
            self.assert_tensor_equal(
                image_center.affine,
                image_mask.affine,
                msg='Physical position is different after cropping',
            )

    def test_mask_corners(self):
        """The mask bounding box and the input image have the same center."""
        target_shape = 8, 22, 30
        transform_center = tio.CropOrPad(target_shape)
        transform_mask = tio.CropOrPad(
            target_shape,
            mask_name='label',
        )
        mask = self.sample_subject['label'].data
        mask *= 0
        mask[0, 0, 0, 0] = 1
        mask[0, -1, -1, -1] = 1
        transformed_center = transform_center(self.sample_subject)
        transformed_mask = transform_mask(self.sample_subject)
        zipped = zip(transformed_center.values(), transformed_mask.values())
        for image_center, image_mask in zipped:
            self.assert_tensor_equal(
                image_center.data,
                image_mask.data,
                msg='Data is different after cropping',
            )
            self.assert_tensor_equal(
                image_center.affine,
                image_mask.affine,
                msg='Physical position is different after cropping',
            )

    def test_2d(self):
        # https://github.com/TorchIO-project/torchio/issues/434
        image = np.random.rand(1, 16, 16, 1)
        mask = np.zeros_like(image, dtype=bool)
        mask[0, 7, 0] = True
        subject = tio.Subject(
            image=tio.ScalarImage(tensor=image),
            mask=tio.LabelMap(tensor=mask),
        )
        transform = tio.CropOrPad((12, 12, 1), mask_name='mask')
        transformed = transform(subject)
        assert transformed.shape == (1, 12, 12, 1)

    def test_no_target_no_mask(self):
        with pytest.raises(ValueError):
            tio.CropOrPad()

    def test_labels_but_no_mask(self):
        with pytest.raises(ValueError):
            tio.CropOrPad(target_shape=(3, 4, 5), labels=[2, 3])

    def test_no_target(self):
        crop_with_mask = tio.CropOrPad(mask_name='label')
        crop_with_mask(self.sample_subject)

    def test_persistent_bounds_params(self):
        # https://github.com/TorchIO-project/torchio/issues/757
        shape = (1, 5, 5, 5)
        mask_a = np.zeros(shape)
        mask_a[0, 2, 2, 2] = 1
        mask_b = mask_a.copy()
        mask_b[0, 1:4, 1:4, 1:4] = 1
        tensor = np.ones(shape)
        image_a = tio.ScalarImage(tensor=tensor)
        mask_a = tio.LabelMap(tensor=mask_a)
        subject_a = tio.Subject(image=image_a, mask=mask_a)
        image_b = tio.ScalarImage(tensor=tensor)
        mask_b = tio.LabelMap(tensor=mask_b)
        subject_b = tio.Subject(image=image_b, mask=mask_b)
        crop = tio.CropOrPad(mask_name='mask')
        for _ in range(2):
            shape_a = crop(subject_a).image.shape
            shape_b = crop(subject_b).image.shape
            assert shape_a != shape_b

    def test_only_crop_pad_true(self):
        with pytest.raises(ValueError):
            tio.CropOrPad((1, 2, 3), only_crop=True, only_pad=True)

    def test_only_pad_true(self):
        target_shape = 9, 21, 30
        orig_shape = self.sample_subject['t1'].spatial_shape
        expected_shape = tuple(
            t if t > o else o for o, t in zip(orig_shape, target_shape)
        )
        transform = tio.CropOrPad(target_shape, only_pad=True)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert result_shape == expected_shape

    def test_only_crop_true(self):
        target_shape = 9, 21, 30
        orig_shape = self.sample_subject['t1'].spatial_shape
        expected_shape = tuple(
            t if t < o else o for o, t in zip(orig_shape, target_shape)
        )
        transform = tio.CropOrPad(target_shape, only_crop=True)
        transformed = transform(self.sample_subject)
        for key in transformed:
            result_shape = transformed[key].spatial_shape
            assert result_shape == expected_shape



import SimpleITK as sitk
import torch

import torchio as tio
from torchio.data.io import sitk_to_nib

from ...utils import TorchioTestCase


class TestPad(TorchioTestCase):
    """Tests for `Pad`."""

    def test_pad(self):
        image = self.sample_subject.t1
        padding = 1, 2, 3, 4, 5, 6
        sitk_image = image.as_sitk()
        low, high = padding[::2], padding[1::2]
        sitk_padded = sitk.ConstantPad(sitk_image, low, high, 0)
        tio_padded = tio.Pad(padding, padding_mode=0)(image)
        sitk_tensor, sitk_affine = sitk_to_nib(sitk_padded)
        tio_tensor, tio_affine = sitk_to_nib(tio_padded.as_sitk())
        self.assert_tensor_equal(sitk_tensor, tio_tensor)
        self.assert_tensor_equal(sitk_affine, tio_affine)

    def test_nans_history(self):
        padded = tio.Pad(1, padding_mode=2)(self.sample_subject)
        again = padded.history[0](self.sample_subject)
        assert not torch.isnan(again.t1.data).any()

    def test_padding_modes(self):
        def padding_func():
            return

        for padding_mode in [0, *tio.Pad.PADDING_MODES, padding_func]:
            tio.Pad(0, padding_mode=padding_mode)

        with self.assertRaises(KeyError):
            tio.Pad(0, padding_mode='abc')

    def test_padding_mean_label_map(self):
        with self.assertWarns(RuntimeWarning):
            tio.Pad(1, padding_mode='mean')(self.sample_subject.label)



import copy

import pytest
import torch

import torchio as tio

from .utils import TorchioTestCase


class TestUtils(TorchioTestCase):
    """Tests for `utils` module."""

    def test_to_tuple(self):
        assert tio.utils.to_tuple(1) == (1,)
        assert tio.utils.to_tuple((1,)) == (1,)
        assert tio.utils.to_tuple(1, length=3) == (1, 1, 1)
        assert tio.utils.to_tuple((1, 2)) == (1, 2)
        assert tio.utils.to_tuple((1, 2), length=3) == (1, 2)
        assert tio.utils.to_tuple([1, 2], length=3) == (1, 2)

    def test_get_stem(self):
        assert tio.utils.get_stem('/home/image.nii.gz') == 'image'
        assert tio.utils.get_stem('/home/image.nii') == 'image'
        assert tio.utils.get_stem('/home/image.nrrd') == 'image'

    def test_guess_type(self):
        assert tio.utils.guess_type('None') is None
        assert isinstance(tio.utils.guess_type('1'), int)
        assert isinstance(tio.utils.guess_type('1.5'), float)
        assert isinstance(tio.utils.guess_type('(1, 3, 5)'), tuple)
        assert isinstance(tio.utils.guess_type('(1,3,5)'), tuple)
        assert isinstance(tio.utils.guess_type('[1,3,5]'), list)
        assert isinstance(tio.utils.guess_type('test'), str)

    def test_apply_transform_to_file(self):
        transform = tio.RandomFlip()
        tio.utils.apply_transform_to_file(
            self.get_image_path('input'),
            transform,
            self.get_image_path('output'),
            verbose=True,
        )

    def test_subjects_from_batch(self):
        dataset = tio.SubjectsDataset(4 * [self.sample_subject])
        loader = tio.SubjectsLoader(dataset, batch_size=4)
        batch = tio.utils.get_first_item(loader)
        subjects = tio.utils.get_subjects_from_batch(batch)
        assert isinstance(subjects[0], tio.Subject)

    def test_subjects_from_batch_with_string_metadata(self):
        subject_c_with_string_metadata = tio.Subject(
            name='John Doe',
            label=tio.LabelMap(self.get_image_path('label_c', binary=True)),
        )

        dataset = tio.SubjectsDataset(4 * [subject_c_with_string_metadata])
        loader = tio.SubjectsLoader(dataset, batch_size=4)
        batch = tio.utils.get_first_item(loader)
        subjects = tio.utils.get_subjects_from_batch(batch)
        assert isinstance(subjects[0], tio.Subject)
        assert 'label' in subjects[0]
        assert 'name' in subjects[0]

    def test_subjects_from_batch_with_int_metadata(self):
        subject_c_with_int_metadata = tio.Subject(
            age=45,
            label=tio.LabelMap(self.get_image_path('label_c', binary=True)),
        )
        dataset = tio.SubjectsDataset(4 * [subject_c_with_int_metadata])
        loader = tio.SubjectsLoader(dataset, batch_size=4)
        batch = tio.utils.get_first_item(loader)
        subjects = tio.utils.get_subjects_from_batch(batch)
        assert isinstance(subjects[0], tio.Subject)
        assert 'label' in subjects[0]
        assert 'age' in subjects[0]

    def test_add_images_from_batch(self):
        subject = copy.deepcopy(self.sample_subject)
        subjects = 4 * [subject]
        preds = torch.rand(4, *subject.shape)
        tio.utils.add_images_from_batch(subjects, preds)

    def test_empty_batch(self):
        with pytest.raises(RuntimeError):
            tio.utils.get_batch_images_and_size({})



import enum

import SimpleITK as sitk


class Interpolation(enum.Enum):
    """Interpolation techniques available in ITK.

    For a full quantitative comparison of interpolation methods, you can read
    `Meijering et al. 1999, Quantitative Comparison of Sinc-Approximating Kernels for Medical Image Interpolation <https://link.springer.com/chapter/10.1007/10704282_23>`_

    Example:
        >>> import torchio as tio
        >>> transform = tio.RandomAffine(image_interpolation='bspline')
    """

    #: Interpolates image intensity at a non-integer pixel position by copying the intensity for the nearest neighbor.
    NEAREST = 'sitkNearestNeighbor'

    #: Linearly interpolates image intensity at a non-integer pixel position.
    LINEAR = 'sitkLinear'

    #: B-Spline of order 3 (cubic) interpolation.
    BSPLINE = 'sitkBSpline'

    #: Same as ``nearest``.
    CUBIC = 'sitkBSpline'

    #: Gaussian interpolation. Sigma is set to 0.8 input pixels and alpha is 4
    GAUSSIAN = 'sitkGaussian'

    #: Smoothly interpolate multi-label images. Sigma is set to 1 input pixel and alpha is 1
    LABEL_GAUSSIAN = 'sitkLabelGaussian'

    #: Hamming windowed sinc kernel.
    HAMMING = 'sitkHammingWindowedSinc'

    #: Cosine windowed sinc kernel.
    COSINE = 'sitkCosineWindowedSinc'

    #: Welch windowed sinc kernel.
    WELCH = 'sitkWelchWindowedSinc'

    #: Lanczos windowed sinc kernel.
    LANCZOS = 'sitkLanczosWindowedSinc'

    #: Blackman windowed sinc kernel.
    BLACKMAN = 'sitkBlackmanWindowedSinc'


def get_sitk_interpolator(interpolation: str) -> int:
    if not isinstance(interpolation, str):
        message = (
            f'Interpolation must be a string, not "{interpolation}"'
            f' of type {type(interpolation)}'
        )
        raise ValueError(message)
    string = getattr(Interpolation, interpolation.upper()).value
    return getattr(sitk, string)



import pytest

import torchio as tio

from ..utils import TorchioTestCase


class TestIXI(TorchioTestCase):
    """Tests for `ixi` module."""

    def test_not_downloaded(self):
        with pytest.raises(RuntimeError):
            tio.datasets.IXI('testing123', download=False)

    def test_tiny_not_downloaded(self):
        with pytest.raises(RuntimeError):
            tio.datasets.IXITiny('testing123', download=False)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import TypeVar

import numpy as np
import torch
from torch.utils.data import DataLoader
from torch.utils.data import Dataset

from .subject import Subject

T = TypeVar('T')


class SubjectsLoader(DataLoader):
    def __init__(
        self,
        dataset: Dataset,
        collate_fn: Callable[[list[T]], Any] | None = None,
        **kwargs,
    ):
        if collate_fn is None:
            collate_fn = self._collate  # type: ignore[assignment]
        super().__init__(
            dataset=dataset,
            collate_fn=collate_fn,
            **kwargs,
        )

    @staticmethod
    def _collate(subjects: list[Subject]) -> dict[str, Any]:
        first_subject = subjects[0]
        batch_dict = {}
        for key in first_subject.keys():
            collated_value = _stack([subject[key] for subject in subjects])
            batch_dict[key] = collated_value
        return batch_dict


def _stack(x):
    """Determine the type of the input and stack it accordingly.

    Args:
        x: List of elements to stack.
    Returns:
        Stacked elements, as either a torch.Tensor, np.ndarray, dict or list.
    """
    first_element = x[0]
    if isinstance(first_element, torch.Tensor):
        return torch.stack(x, dim=0)
    elif isinstance(first_element, np.ndarray):
        return np.stack(x, axis=0)
    elif isinstance(first_element, dict):
        # Assume that all elements have the same keys
        collated_dict = {}
        for key in first_element.keys():
            collated_dict[key] = _stack([element[key] for element in x])
        return collated_dict
    else:
        return x



import pytest
import torch

from torchio import LABEL
from torchio.transforms import Lambda

from ..utils import TorchioTestCase


class TestLambda(TorchioTestCase):
    """Tests for :class:`Lambda` class."""

    def test_wrong_return_type(self):
        transform = Lambda(lambda x: 'Not a tensor')
        with pytest.raises(ValueError):
            transform(self.sample_subject)

    def test_wrong_return_data_type(self):
        transform = Lambda(lambda x: torch.rand(1) > 0)
        with pytest.raises(ValueError):
            transform(self.sample_subject)

    def test_wrong_return_shape(self):
        transform = Lambda(lambda x: torch.rand(1))
        with pytest.raises(ValueError):
            transform(self.sample_subject)

    def test_lambda(self):
        transform = Lambda(lambda x: x + 1)
        transformed = transform(self.sample_subject)
        assert torch.all(
            torch.eq(
                transformed.t1.data,
                self.sample_subject.t1.data + 1,
            ),
        )
        assert torch.all(
            torch.eq(
                transformed.t2.data,
                self.sample_subject.t2.data + 1,
            ),
        )
        assert torch.all(
            torch.eq(
                transformed.label.data,
                self.sample_subject.label.data + 1,
            ),
        )

    def test_image_types(self):
        transform = Lambda(lambda x: x + 1, types_to_apply=[LABEL])
        transformed = transform(self.sample_subject)
        assert torch.all(
            torch.eq(
                transformed.t1.data,
                self.sample_subject.t1.data,
            ),
        )
        assert torch.all(
            torch.eq(
                transformed.t2.data,
                self.sample_subject.t2.data,
            ),
        )
        assert torch.all(
            torch.eq(
                transformed.label.data,
                self.sample_subject.label.data + 1,
            ),
        )



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestAggregator(TorchioTestCase):
    """Tests for `aggregator` module."""

    def aggregate(self, mode, fixture):
        image_shape = 1, 1, 4, 4
        tensor = torch.ones(image_shape)
        fixture = torch.as_tensor(fixture).reshape(image_shape)
        image_name = 'img'
        subject = tio.Subject({image_name: tio.ScalarImage(tensor=tensor)})
        patch_size = 1, 3, 3
        patch_overlap = 0, 2, 2
        sampler = tio.data.GridSampler(subject, patch_size, patch_overlap)
        aggregator = tio.data.GridAggregator(sampler, overlap_mode=mode)
        loader = tio.SubjectsLoader(sampler, batch_size=3)
        values_dict = {
            (0, 0): 0,
            (0, 1): 2,
            (1, 0): 4,
            (1, 1): 6,
        }
        for batch in loader:
            iterable = zip(batch[tio.LOCATION], batch[image_name][tio.DATA])
            for location, data in iterable:
                coords_2d = tuple(location[1:3].tolist())
                data *= values_dict[coords_2d]
            batch_data = batch[image_name][tio.DATA]
            aggregator.add_batch(batch_data, batch[tio.LOCATION])
        output = aggregator.get_output_tensor()
        self.assert_tensor_equal(output, fixture)

    def test_overlap_crop(self):
        fixture = (
            (0, 0, 2, 2),
            (0, 0, 2, 2),
            (4, 4, 6, 6),
            (4, 4, 6, 6),
        )
        self.aggregate('crop', fixture)

    def test_overlap_average(self):
        fixture = (
            (0, 1, 1, 2),
            (2, 3, 3, 4),
            (2, 3, 3, 4),
            (4, 5, 5, 6),
        )
        self.aggregate('average', fixture)

    def test_overlap_hann(self):
        fixture = (
            (0 / 3, 2 / 3, 4 / 3, 6 / 3),  # noqa: E201, E241
            (4 / 3, 6 / 3, 8 / 3, 10 / 3),  # noqa: E201, E241
            (8 / 3, 10 / 3, 12 / 3, 14 / 3),  # noqa: E201, E241
            (12 / 3, 14 / 3, 16 / 3, 18 / 3),
        )
        self.aggregate('hann', fixture)

    def run_sampler_aggregator(self, overlap_mode='crop'):
        patch_size = 10
        patch_overlap = 2
        grid_sampler = tio.inference.GridSampler(
            self.sample_subject,
            patch_size,
            patch_overlap,
        )
        patch_loader = tio.SubjectsLoader(grid_sampler)
        aggregator = tio.inference.GridAggregator(
            grid_sampler,
            overlap_mode=overlap_mode,
        )
        for batch in patch_loader:
            data = batch['t1'][tio.DATA].long()
            aggregator.add_batch(data, batch[tio.LOCATION])
        return aggregator

    def test_warning_int64(self):
        aggregator = self.run_sampler_aggregator()
        with pytest.warns(RuntimeWarning):
            aggregator.get_output_tensor()

    def run_patch_crop_issue(self, *, padding_mode):
        # https://github.com/TorchIO-project/torchio/issues/813
        pao, pas, ims, bb1, bb2 = 4, 102, 320, 100, 120

        patch_overlap = pao, 0, 0
        patch_size = pas, 1, 1
        img = torch.zeros((1, ims, 1, 1))
        bbox = [bb1, bb2]

        img[:, bbox[0] : bbox[1]] = 1
        image = tio.LabelMap(tensor=img)
        subject = tio.Subject(image=image)
        grid_sampler = tio.inference.GridSampler(
            subject,
            patch_size,
            patch_overlap,
        )
        patch_loader = tio.SubjectsLoader(grid_sampler)
        aggregator = tio.inference.GridAggregator(grid_sampler)
        for patches_batch in patch_loader:
            input_tensor = patches_batch['image'][tio.DATA]
            locations = patches_batch[tio.LOCATION]
            aggregator.add_batch(input_tensor, locations)
        output_tensor = aggregator.get_output_tensor()
        self.assert_tensor_equal(image.tensor, output_tensor)

    def test_patch_crop_issue_no_padding(self):
        self.run_patch_crop_issue(padding_mode=None)

    def test_patch_crop_issue_padding(self):
        self.run_patch_crop_issue(padding_mode='constant')

    def test_bad_aggregator_shape(self):
        # https://github.com/microsoft/InnerEye-DeepLearning/pull/677/checks?check_run_id=5395915817
        tensor = torch.ones(1, 40, 40, 40)
        image_name = 'img'
        subject = tio.Subject({image_name: tio.ScalarImage(tensor=tensor)})
        patch_size = 40
        patch_overlap = 30
        sampler = tio.data.GridSampler(
            subject,
            patch_size,
            patch_overlap,
            padding_mode='edge',
        )
        aggregator = tio.data.GridAggregator(sampler)
        loader = tio.SubjectsLoader(sampler, batch_size=3)
        for batch in loader:
            input_batch = batch[image_name][tio.DATA]
            crop = tio.CropOrPad(12)
            patches = [crop(patch) for patch in input_batch]
            inference_batch = torch.stack(patches)
            with pytest.raises(RuntimeError):
                aggregator.add_batch(inference_batch, batch[tio.LOCATION])



from collections import defaultdict
from numbers import Number
from typing import Union

import numpy as np
import torch

from ....data.subject import Subject
from ...fourier import FourierTransform
from ...intensity_transform import IntensityTransform
from .. import RandomTransform


class RandomSpike(RandomTransform, IntensityTransform, FourierTransform):
    r"""Add random MRI spike artifacts.

    Also known as `Herringbone artifact
    <https://radiopaedia.org/articles/herringbone-artifact?lang=gb>`_,
    crisscross artifact or corduroy artifact, it creates stripes in different
    directions in image space due to spikes in k-space.

    Args:
        num_spikes: Number of spikes :math:`n` present in k-space.
            If a tuple :math:`(a, b)` is provided, then
            :math:`n \sim \mathcal{U}(a, b) \cap \mathbb{N}`.
            If only one value :math:`d` is provided,
            :math:`n \sim \mathcal{U}(0, d) \cap \mathbb{N}`.
            Larger values generate more distorted images.
        intensity: Ratio :math:`r` between the spike intensity and the maximum
            of the spectrum.
            If a tuple :math:`(a, b)` is provided, then
            :math:`r \sim \mathcal{U}(a, b)`.
            If only one value :math:`d` is provided,
            :math:`r \sim \mathcal{U}(-d, d)`.
            Larger values generate more distorted images.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    .. note:: The execution time of this transform does not depend on the
        number of spikes.
    """

    def __init__(
        self,
        num_spikes: Union[int, tuple[int, int]] = 1,
        intensity: Union[float, tuple[float, float]] = (1, 3),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.intensity_range = self._parse_range(
            intensity,
            'intensity_range',
        )
        self.num_spikes_range: tuple[int, int] = self._parse_range(  # type: ignore[assignment]
            num_spikes,
            'num_spikes',
            min_constraint=0,
            type_constraint=int,
        )

    def apply_transform(self, subject: Subject) -> Subject:
        images_dict = self.get_images_dict(subject)
        if not images_dict:
            return subject

        arguments: dict[str, dict] = defaultdict(dict)
        for image_name in images_dict:
            spikes_positions_param, intensity_param = self.get_params(
                self.num_spikes_range,
                self.intensity_range,
            )
            arguments['spikes_positions'][image_name] = spikes_positions_param
            arguments['intensity'][image_name] = intensity_param
        transform = Spike(**self.add_base_args(arguments))
        transformed = transform(subject)
        assert isinstance(transformed, Subject)
        return transformed

    def get_params(
        self,
        num_spikes_range: tuple[int, int],
        intensity_range: tuple[float, float],
    ) -> tuple[np.ndarray, float]:
        ns_min, ns_max = num_spikes_range
        num_spikes_param = int(torch.randint(ns_min, ns_max + 1, (1,)).item())
        intensity_param = self.sample_uniform(*intensity_range)
        spikes_positions = torch.rand(num_spikes_param, 3).numpy()
        return spikes_positions, intensity_param


class Spike(IntensityTransform, FourierTransform):
    r"""Add MRI spike artifacts.

    Also known as `Herringbone artifact
    <https://radiopaedia.org/articles/herringbone-artifact>`_,
    crisscross artifact or corduroy artifact, it creates stripes in different
    directions in image space due to spikes in k-space.

    Args:
        spikes_positions:
        intensity: Ratio :math:`r` between the spike intensity and the maximum
            of the spectrum.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    .. note:: The execution time of this transform does not depend on the
        number of spikes.
    """

    def __init__(
        self,
        spikes_positions: Union[np.ndarray, dict[str, np.ndarray]],
        intensity: Union[float, dict[str, float]],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.spikes_positions = spikes_positions
        self.intensity = intensity
        self.args_names = ['spikes_positions', 'intensity']
        self.invert_transform = False

    def apply_transform(self, subject: Subject) -> Subject:
        spikes_positions = self.spikes_positions
        intensity = self.intensity
        for image_name, image in self.get_images_dict(subject).items():
            if self.arguments_are_dict():
                spikes_positions = self.spikes_positions[image_name]
                assert isinstance(self.intensity, dict)
                intensity = self.intensity[image_name]
            transformed_tensors = []
            for channel in image.data:
                assert isinstance(intensity, Number)
                transformed_tensor = self.add_artifact(
                    channel,
                    np.asarray(spikes_positions),
                    intensity,
                )
                transformed_tensors.append(transformed_tensor)
            image.set_data(torch.stack(transformed_tensors))
        return subject

    def add_artifact(
        self,
        tensor: torch.Tensor,
        spikes_positions: np.ndarray,
        intensity_factor: float,
    ):
        if intensity_factor == 0 or len(spikes_positions) == 0:
            return tensor
        spectrum = self.fourier_transform(tensor)
        shape = np.array(spectrum.shape)
        mid_shape = shape // 2
        indices = np.floor(spikes_positions * shape).astype(int)
        for index in indices:
            diff = index - mid_shape
            i, j, k = mid_shape + diff
            # As of torch 1.7, "max is not yet implemented for complex tensors"
            artifact = spectrum.cpu().numpy().max() * intensity_factor
            if self.invert_transform:
                spectrum[i, j, k] -= artifact
            else:
                spectrum[i, j, k] += artifact
            # If we wanted to add a pure cosine, we should add spikes to both
            # sides of k-space. However, having only one is a better
            # representation og the actual cause of the artifact in real
            # scans. Therefore the next two lines have been removed.
            # #i, j, k = mid_shape - diff
            # #spectrum[i, j, k] = spectrum.max() * intensity_factor
        result = self.inv_fourier_transform(spectrum).real.float()
        return result



import numpy as np
import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestResample(TorchioTestCase):
    """Tests for `Resample`."""

    def test_spacing(self):
        # Should this raise an error if sizes are different?
        spacing = 2
        transform = tio.Resample(spacing)
        transformed = transform(self.sample_subject)
        for image in transformed.get_images(intensity_only=False):
            assert image.spacing == 3 * (spacing,)

    def test_reference_name(self):
        subject = self.get_inconsistent_shape_subject()
        reference_name = 't1'
        transform = tio.Resample(reference_name)
        transformed = transform(subject)
        reference_image = subject[reference_name]
        for image in transformed.get_images(intensity_only=False):
            assert reference_image.shape == image.shape
            self.assert_tensor_almost_equal(
                reference_image.affine,
                image.affine,
            )

    def test_affine(self):
        spacing = 1
        affine_name = 'pre_affine'
        transform = tio.Resample(spacing, pre_affine_name=affine_name)
        transformed = transform(self.sample_subject)
        for image in transformed.values():
            if affine_name in image:
                target_affine = np.eye(4)
                target_affine[:3, 3] = 10, 0, -0.1
                self.assert_tensor_almost_equal(image.affine, target_affine)
            else:
                self.assert_tensor_equal(image.affine, np.eye(4))

    def test_missing_affine(self):
        transform = tio.Resample(1, pre_affine_name='missing')
        with pytest.raises(ValueError):
            transform(self.sample_subject)

    def test_reference_path(self):
        reference_image, reference_path = self.get_reference_image_and_path()
        transform = tio.Resample(reference_path)
        transformed = transform(self.sample_subject)
        for image in transformed.values():
            assert reference_image.shape == image.shape
            self.assert_tensor_almost_equal(
                reference_image.affine,
                image.affine,
            )

    def test_wrong_spacing_length(self):
        with pytest.raises(RuntimeError):
            tio.Resample((1, 2))(self.sample_subject)

    def test_wrong_spacing_value(self):
        with pytest.raises(ValueError):
            tio.Resample(0)(self.sample_subject)

    def test_wrong_target_type(self):
        with pytest.raises(RuntimeError):
            tio.Resample(None)(self.sample_subject)

    def test_missing_reference(self):
        transform = tio.Resample('missing')
        with pytest.raises(ValueError):
            transform(self.sample_subject)

    def test_2d(self):
        image = tio.ScalarImage(tensor=torch.rand(1, 2, 3, 1))
        transform = tio.Resample(0.5)
        shape = transform(image).shape
        assert shape == (1, 4, 6, 1)

    def test_input_list(self):
        tio.Resample([1, 2, 3])(self.sample_subject)

    def test_input_array(self):
        resample = tio.Resample(np.asarray([1, 2, 3]))
        resample(self.sample_subject)

    def test_image_target(self):
        tio.Resample(self.sample_subject.t1)(self.sample_subject)

    def test_bad_affine(self):
        shape = 1, 2, 3
        affine = np.eye(3)
        target = shape, affine
        transform = tio.Resample(target)
        with pytest.raises(RuntimeError):
            transform(self.sample_subject)



#!/usr/bin/env python
from copy import copy

import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestGridSampler(TorchioTestCase):
    """Tests for `GridSampler`."""

    def test_locations(self):
        patch_size = 5, 20, 20
        patch_overlap = 2, 4, 6
        sampler = tio.GridSampler(
            subject=self.sample_subject,
            patch_size=patch_size,
            patch_overlap=patch_overlap,
        )
        fixture = [
            [0, 0, 0, 5, 20, 20],
            [0, 0, 10, 5, 20, 30],
            [3, 0, 0, 8, 20, 20],
            [3, 0, 10, 8, 20, 30],
            [5, 0, 0, 10, 20, 20],
            [5, 0, 10, 10, 20, 30],
        ]
        locations = sampler.locations.tolist()
        assert locations == fixture

    def test_generate_patches(self):
        patch_size = 5, 15, 15
        sampler = tio.GridSampler(self.sample_subject, patch_size)
        for patch in sampler():
            assert patch.spatial_shape == patch_size

    def test_large_patch(self):
        with pytest.raises(ValueError):
            tio.GridSampler(self.sample_subject, (5, 21, 5), (0, 2, 0))

    def test_large_overlap(self):
        with pytest.raises(ValueError):
            tio.GridSampler(self.sample_subject, (5, 20, 5), (2, 4, 6))

    def test_odd_overlap(self):
        with pytest.raises(ValueError):
            tio.GridSampler(self.sample_subject, (5, 20, 5), (2, 4, 3))

    def test_single_location(self):
        sampler = tio.GridSampler(self.sample_subject, (10, 20, 30), 0)
        fixture = [[0, 0, 0, 10, 20, 30]]
        assert sampler.locations.tolist() == fixture

    def test_subject_shape(self):
        patch_size = 5, 20, 20
        patch_overlap = 2, 4, 6
        initial_shape = copy(self.sample_subject.shape)
        tio.GridSampler(
            self.sample_subject,
            patch_size,
            patch_overlap,
            padding_mode='reflect',
        )
        final_shape = self.sample_subject.shape
        assert initial_shape == final_shape



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestContour(TorchioTestCase):
    """Tests for `Contour`."""

    def test_one_hot(self):
        image = self.sample_subject.label
        tio.Contour()(image)

    def test_multichannel(self):
        label_map = tio.LabelMap(tensor=torch.rand(2, 3, 3, 3) > 1)
        with pytest.raises(RuntimeError):
            tio.Contour()(label_map)



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestMask(TorchioTestCase):
    def test_single_mask(self):
        negated_mask = self.sample_subject.label.data.logical_not()
        masked_voxel_indices = negated_mask.nonzero(as_tuple=True)
        transform = tio.Mask(masking_method='label')
        transformed = transform(self.sample_subject)
        assert (transformed.t1.data[masked_voxel_indices] == 0).all()

    def test_single_mask_nonzero_background(self):
        background_value = 314159
        negated_mask = self.sample_subject.label.data.logical_not()
        masked_voxel_indices = negated_mask.nonzero(as_tuple=True)

        transform = tio.Mask(
            masking_method='label',
            outside_value=background_value,
        )
        transformed = transform(self.sample_subject)

        assert (transformed.t1.data[masked_voxel_indices] == background_value).all()

    def test_mask_specified_label(self):
        mask_label = [1]
        negated_mask = self.sample_subject.label.data.logical_not()
        masked_voxel_indices = negated_mask.nonzero(as_tuple=True)

        transform = tio.Mask(masking_method='label', labels=mask_label)
        transformed = transform(self.sample_subject)

        assert (transformed.t1.data[masked_voxel_indices] == 0).all()

    def test_mask_specified_label_small(self):
        def to_image(*numbers):
            return torch.as_tensor(numbers).reshape(1, 1, 1, len(numbers))

        image_tensor = to_image(1, 6, 7, 3, 0)
        label_tensor = to_image(0, 1, 2, 3, 4)
        mask_labels = [1, 2]
        subject = tio.Subject(
            image=tio.ScalarImage(tensor=image_tensor),
            label=tio.LabelMap(tensor=label_tensor),
        )
        transform = tio.Mask(masking_method='label', labels=mask_labels)
        transformed = transform(subject)
        masked_list = transformed.image.data.flatten().tolist()
        assert masked_list == [0, 6, 7, 0, 0]

    def test_mask_example(self):
        subject = self.sample_subject
        negated_mask = subject.label.data.logical_not()
        masked_voxel_indices = negated_mask.nonzero(as_tuple=True)
        transform = tio.Mask(masking_method='label')
        transformed = transform(subject)
        assert (transformed.t1.data[masked_voxel_indices] == 0).all()

    def test_4d(self):
        image = tio.ScalarImage(tensor=torch.rand(3, 4, 5, 6))
        mask = tio.LabelMap(tensor=torch.ones(1, 4, 5, 6))
        subject = tio.Subject(image=image, mask_lm=mask)
        transform = tio.Mask(masking_method='mask_lm')
        with pytest.warns(RuntimeWarning, match='^Expanding.*'):
            masked = transform(subject)
        assert masked.image.shape == image.shape



import pytest
import torch

import torchio as tio

from ..utils import TorchioTestCase


class TestSubjectsDataset(TorchioTestCase):
    def test_indexing_nonint(self):
        dset = tio.SubjectsDataset(self.subjects_list)
        dset[torch.tensor(0)]

    def test_images(self):
        self.iterate_dataset(self.subjects_list)

    def test_empty_subjects_list(self):
        with pytest.raises(ValueError):
            self.iterate_dataset([])

    def test_empty_subjects_tuple(self):
        with pytest.raises(ValueError):
            self.iterate_dataset(())

    def test_wrong_subjects_type(self):
        with pytest.raises(TypeError):
            self.iterate_dataset(0)

    def test_wrong_subject_type_int(self):
        with pytest.raises(TypeError):
            self.iterate_dataset([0])

    def test_wrong_subject_type_dict(self):
        with pytest.raises(TypeError):
            self.iterate_dataset([{}])

    def test_wrong_index(self):
        with pytest.raises(ValueError):
            self.dataset[:3]

    def test_wrong_transform_init(self):
        with pytest.raises(ValueError):
            tio.SubjectsDataset(
                self.subjects_list,
                transform={},
            )

    def test_wrong_transform_arg(self):
        with pytest.raises(ValueError):
            self.dataset.set_transform(1)

    @staticmethod
    def iterate_dataset(subjects_list):
        dataset = tio.SubjectsDataset(subjects_list)
        for _ in dataset:
            pass

    def test_from_batch(self):
        dataset = tio.SubjectsDataset([self.sample_subject])
        loader = tio.SubjectsLoader(dataset)
        batch = tio.utils.get_first_item(loader)
        new_dataset = tio.SubjectsDataset.from_batch(batch)
        self.assert_tensor_equal(
            dataset[0].t1.data,
            new_dataset[0].t1.data,
        )



from __future__ import annotations

import torch

from ....data.subject import Subject
from .normalization_transform import NormalizationTransform
from .normalization_transform import TypeMaskingMethod


class ZNormalization(NormalizationTransform):
    """Subtract mean and divide by standard deviation.

    Args:
        masking_method: See
            :class:`~torchio.transforms.preprocessing.intensity.NormalizationTransform`.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.
    """

    def __init__(self, masking_method: TypeMaskingMethod = None, **kwargs):
        super().__init__(masking_method=masking_method, **kwargs)
        self.args_names = ['masking_method']

    def apply_normalization(
        self,
        subject: Subject,
        image_name: str,
        mask: torch.Tensor,
    ) -> None:
        image = subject[image_name]
        standardized = self.znorm(
            image.data,
            mask,
        )
        if standardized is None:
            message = (
                'Standard deviation is 0 for masked values'
                f' in image "{image_name}" ({image.path})'
            )
            raise RuntimeError(message)
        image.set_data(standardized)

    @staticmethod
    def znorm(
        tensor: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor | None:
        tensor = tensor.clone().float()
        values = tensor[mask]
        mean, std = values.mean(), values.std()
        if std == 0:
            return None
        tensor -= mean
        tensor /= std
        return tensor



import torch.nn.functional as F  # noqa: N812

from ....data.image import Image
from .label_transform import LabelTransform


class OneHot(LabelTransform):
    r"""Reencode label maps using one-hot encoding.

    Args:
        num_classes: See :func:`~torch.nn.functional.one_hot`.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.
    """

    def __init__(self, num_classes: int = -1, **kwargs):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.args_names = ['num_classes']
        self.invert_transform = False

    def apply_transform(self, subject):
        for image in self.get_images(subject):
            if self.invert_transform:
                self.argmax(image)
            else:
                self.one_hot(image)
        return subject

    @staticmethod
    def argmax(image: Image) -> None:
        data = image.data.argmax(dim=0, keepdim=True)
        image.set_data(data)

    def one_hot(self, image: Image) -> None:
        if image.num_channels > 1:
            message = (
                'The number of input channels must be 1,'
                f' but it is {image.num_channels}'
            )
            raise RuntimeError(message)
        data = image.data[0]
        num_classes = -1 if self.num_classes is None else self.num_classes
        one_hot = F.one_hot(data.long(), num_classes=num_classes)
        image.set_data(one_hot.permute(3, 0, 1, 2).type(data.type()))



from __future__ import annotations

import csv
from pathlib import Path

from ..data import LabelMap
from ..data import ScalarImage
from ..data import Subject
from ..data import SubjectsDataset
from ..download import download_and_extract_archive
from ..transforms import Transform
from ..types import TypePath


class EPISURG(SubjectsDataset):
    """
    `EPISURG <https://doi.org/10.5522/04/9996158.v1>`_ is a clinical dataset of
    :math:`T_1`-weighted MRI from 430 epileptic patients who underwent
    resective brain surgery at the National Hospital of Neurology and
    Neurosurgery (Queen Square, London, United Kingdom) between 1990 and 2018.

    The dataset comprises 430 postoperative MRI. The corresponding preoperative
    MRI is present for 268 subjects.

    Three human raters segmented the resection cavity on partially overlapping
    subsets of EPISURG.

    If you use this dataset for your research, you agree with the *Data use
    agreement* presented at the EPISURG entry on the `UCL Research Data
    Repository <https://doi.org/10.5522/04/9996158.v1>`_ and you must cite the
    corresponding publications.

    Args:
        root: Root directory to which the dataset will be downloaded.
        transform: An instance of
            :class:`~torchio.transforms.transform.Transform`.
        download: If set to ``True``, will download the data into :attr:`root`.

    .. warning:: The size of this dataset is multiple GB.
        If you set :attr:`download` to ``True``, it will take some time
        to be downloaded if it is not already present.
    """

    data_url = 'https://s3-eu-west-1.amazonaws.com/pstorage-ucl-2748466690/26153588/EPISURG.zip'
    md5 = '5ec5831a2c6fbfdc8489ba2910a6504b'

    def __init__(
        self,
        root: TypePath,
        transform: Transform | None = None,
        download: bool = False,
        **kwargs,
    ):
        root = Path(root).expanduser().absolute()
        if download:
            self._download(root)
        subjects_list = self._get_subjects_list(root)
        self.kwargs = kwargs
        super().__init__(subjects_list, transform=transform, **kwargs)

    @staticmethod
    def _check_exists(root, modalities):
        for modality in modalities:
            modality_dir = root / modality
            if not modality_dir.is_dir():
                exists = False
                break
        else:
            exists = True
        return exists

    @staticmethod
    def _get_subjects_list(root):
        episurg_dir = root / 'EPISURG'
        subjects_dir = episurg_dir / 'subjects'
        csv_path = episurg_dir / 'subjects.csv'
        with open(csv_path) as csvfile:
            reader = csv.DictReader(csvfile)
            subjects = []
            for row in reader:
                subject_id = row['Subject']
                subject_dir = subjects_dir / subject_id
                subject_dict = {
                    'subject_id': subject_id,
                    'hemisphere': row['Hemisphere'],
                    'surgery_type': row['Type'],
                }
                preop_dir = subject_dir / 'preop'
                preop_paths = list(preop_dir.glob('*preop*'))
                assert len(preop_paths) <= 1
                if preop_paths:
                    subject_dict['preop_mri'] = ScalarImage(preop_paths[0])
                postop_dir = subject_dir / 'postop'
                postop_path = list(postop_dir.glob('*postop-t1mri*'))[0]
                subject_dict['postop_mri'] = ScalarImage(postop_path)
                for seg_path in postop_dir.glob('*seg*'):
                    seg_id = seg_path.name[-8]
                    subject_dict[f'seg_{seg_id}'] = LabelMap(seg_path)
                subjects.append(Subject(**subject_dict))
        return subjects

    def _download(self, root: Path):
        """Download the EPISURG data if it does not exist already."""
        if (root / 'EPISURG').is_dir():
            return
        root.mkdir(exist_ok=True, parents=True)
        download_and_extract_archive(
            self.data_url,
            download_root=root,
            md5=self.md5,
        )
        (root / 'EPISURG.zip').unlink()  # cleanup

    def _glob_subjects(self, string):
        subjects = []
        for subject in self._subjects:
            for image_name in subject:
                if string in image_name:
                    subjects.append(subject)
                    break
        return subjects

    def _get_labeled_subjects(self):
        return self._glob_subjects('seg')

    def _get_paired_subjects(self):
        return self._glob_subjects('preop')

    def _get_subset(self, subjects):
        dataset = SubjectsDataset(
            subjects,
            transform=self._transform,
            **(self.kwargs),
        )
        return dataset

    def get_labeled(self) -> SubjectsDataset:
        """Get dataset from subjects with manual annotations."""
        return self._get_subset(self._get_labeled_subjects())

    def get_unlabeled(self) -> SubjectsDataset:
        """Get dataset from subjects without manual annotations."""
        subjects = [s for s in self._subjects if s not in self._get_labeled_subjects()]
        return self._get_subset(subjects)

    def get_paired(self) -> SubjectsDataset:
        """Get dataset from subjects with pre- and post-op MRI."""
        return self._get_subset(self._get_paired_subjects())



"""Top-level package for torchio."""

__author__ = """Fernando Perez-Garcia"""
__email__ = 'fepegar@gmail.com'
__version__ = '0.20.10'


from . import datasets
from . import reference
from . import utils
from .constants import *  # noqa: F401, F403
from .data import GridAggregator
from .data import GridSampler
from .data import Image
from .data import LabelMap
from .data import LabelSampler
from .data import Queue
from .data import ScalarImage
from .data import Subject
from .data import SubjectsDataset
from .data import SubjectsLoader
from .data import UniformSampler
from .data import WeightedSampler
from .data import inference
from .data import io
from .data import sampler
from .transforms import *  # noqa: F401, F403

__all__ = [
    'utils',
    'io',
    'sampler',
    'inference',
    'SubjectsDataset',
    'SubjectsLoader',
    'Image',
    'ScalarImage',
    'LabelMap',
    'Queue',
    'Subject',
    'datasets',
    'reference',
    'WeightedSampler',
    'UniformSampler',
    'LabelSampler',
    'GridSampler',
    'GridAggregator',
]



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestClamp(TorchioTestCase):
    """Tests for :class:`tio.Clamp` class."""

    def test_out_min_max(self):
        transform = tio.Clamp(out_min=0, out_max=1)
        transformed = transform(self.sample_subject)
        assert transformed.t1.data.min() == 0
        assert transformed.t1.data.max() == 1

    def test_ct(self):
        ct_max = 1500
        ct_min = -2000
        ct_range = ct_max - ct_min
        tensor = torch.rand(1, 30, 30, 30) * ct_range + ct_min
        ct = tio.ScalarImage(tensor=tensor)
        ct_air = -1000
        ct_bone = 1000
        clamp = tio.Clamp(ct_air, ct_bone)
        clamped = clamp(ct)
        assert clamped.data.min() == ct_air
        assert clamped.data.max() == ct_bone

    def test_too_many_values_for_out_min(self):
        with pytest.raises(TypeError):
            clamp = tio.Clamp(out_min=(1, 2))
            clamp(self.sample_subject)

    def test_too_many_values_for_out_max(self):
        with pytest.raises(TypeError):
            clamp = tio.Clamp(out_max=(1, 2))
            clamp(self.sample_subject)

    def test_wrong_out_min_type(self):
        with pytest.raises(TypeError):
            clamp = tio.Clamp(out_min='foo')
            clamp(self.sample_subject)

    def test_wrong_out_max_type(self):
        with pytest.raises(TypeError):
            clamp = tio.Clamp(out_max='foo')
            clamp(self.sample_subject)



from __future__ import annotations

import abc
from pathlib import Path

from ..data import Image
from ..data import LabelMap
from ..data import ScalarImage
from ..data import Subject
from ..data import SubjectsDataset
from ..download import download_and_extract_archive
from ..transforms import Transform
from ..types import TypePath


class BITE(SubjectsDataset, abc.ABC):
    base_url = 'http://www.bic.mni.mcgill.ca/uploads/Services/'

    def __init__(
        self,
        root: TypePath,
        transform: Transform | None = None,
        download: bool = False,
        **kwargs,
    ):
        root = Path(root).expanduser().absolute()
        if download:
            self._download(root)
        subjects_list = self._get_subjects_list(root)
        self.kwargs = kwargs
        super().__init__(subjects_list, transform=transform, **kwargs)

    @abc.abstractmethod
    def _download(self, root: Path):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_subjects_list(self, root: Path):
        raise NotImplementedError


class BITE3(BITE):
    """Pre- and post-resection MR images in BITE.

    *The goal of BITE is to share in vivo medical images of patients wtith
    brain tumors to facilitate the development and validation of new image
    processing algorithms.*

    Please check the `BITE website`_ for more information and
    acknowledgments instructions.

    .. _BITE website: https://nist.mni.mcgill.ca/bite-brain-images-of-tumors-for-evaluation-database/

    Args:
        root: Root directory to which the dataset will be downloaded.
        transform: An instance of
            :class:`~torchio.transforms.transform.Transform`.
        download: If set to ``True``, will download the data into :attr:`root`.
    """

    dirname = 'group3'

    def _download(self, root: Path):
        if (root / self.dirname).is_dir():
            return
        root.mkdir(exist_ok=True, parents=True)
        filename = f'{self.dirname}.tar.gz'
        url = self.base_url + filename
        download_and_extract_archive(
            url,
            download_root=root,
            md5='e415b63887c40b727c45552614b44634',
        )
        (root / filename).unlink()  # cleanup

    def _get_subjects_list(self, root: Path):
        subjects_dir = root / self.dirname
        subjects = []
        for i in range(1, 15):
            if i == 13:
                continue  # no MRI for this subject
            subject_id = f'{i:02d}'
            subject_dir = subjects_dir / subject_id
            preop_path = subject_dir / f'{subject_id}_preop_mri.mnc'
            postop_path = subject_dir / f'{subject_id}_postop_mri.mnc'
            images_dict: dict[str, Image] = {}
            images_dict['preop'] = ScalarImage(preop_path)
            images_dict['postop'] = ScalarImage(postop_path)
            for fp in subject_dir.glob('*tumor*'):
                images_dict[fp.stem[3:]] = LabelMap(fp)
            subject = Subject(images_dict)
            subjects.append(subject)
        return subjects



from ...data.subject import Subject
from ...utils import get_torchio_cache_dir


class SubjectMNI(Subject):
    """Atlases from the Montreal Neurological Institute (MNI).

    See `the website <https://nist.mni.mcgill.ca/?page_id=714>`_ for more
    information.
    """

    url_base = 'http://packages.bic.mni.mcgill.ca/mni-models/'

    @property
    def download_root(self):
        return get_torchio_cache_dir() / self.name



from __future__ import annotations

from collections.abc import Sequence

import torch

from ..constants import TYPE
from ..data.subject import Subject
from ..types import TypeCallable
from .transform import Transform


class Lambda(Transform):
    """Applies a user-defined function as transform.

    Args:
        function: Callable that receives and returns a 4D
            :class:`torch.Tensor`.
        types_to_apply: List of strings corresponding to the image types to
            which this transform should be applied. If ``None``, the transform
            will be applied to all images in the subject.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.

    Example:
        >>> import torchio as tio
        >>> invert_intensity = tio.Lambda(lambda x: -x, types_to_apply=[tio.INTENSITY])
        >>> invert_mask = tio.Lambda(lambda x: 1 - x, types_to_apply=[tio.LABEL])
        >>> def double(x):
        ...     return 2 * x
        >>> double_transform = tio.Lambda(double)
    """

    def __init__(
        self,
        function: TypeCallable,
        types_to_apply: Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.function = function
        self.types_to_apply = types_to_apply
        self.args_names = ['function', 'types_to_apply']

    def apply_transform(self, subject: Subject) -> Subject:
        images = subject.get_images(
            intensity_only=False,
            include=self.include,
            exclude=self.exclude,
        )
        for image in images:
            image_type = image[TYPE]
            if self.types_to_apply is not None:
                if image_type not in self.types_to_apply:
                    continue

            function_arg = image.data
            result = self.function(function_arg)
            if not isinstance(result, torch.Tensor):
                message = (
                    'The returned value from the callable argument must be'
                    f' of type {torch.Tensor}, not {type(result)}'
                )
                raise ValueError(message)
            if result.ndim != function_arg.ndim:
                message = (
                    'The number of dimensions of the returned value must'
                    f' be {function_arg.ndim}, not {result.ndim}'
                )
                raise ValueError(message)
            image.set_data(result)
        return subject



# pylint: disable=import-outside-toplevel

from pathlib import Path

import typer
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TextColumn

app = typer.Typer()


@app.command()
def main(
    input_path: Path = typer.Argument(  # noqa: B008
        ...,
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
    ),
    transform_name: str = typer.Argument(...),  # noqa: B008
    output_path: Path = typer.Argument(  # noqa: B008
        ...,
        file_okay=True,
        dir_okay=False,
        writable=True,
    ),
    kwargs: str = typer.Option(  # noqa: B008
        None,
        '--kwargs',
        '-k',
        help='String of kwargs, e.g. "degrees=(-5,15) num_transforms=3".',
    ),
    imclass: str = typer.Option(  # noqa: B008
        'ScalarImage',
        '--imclass',
        '-c',
        help=(
            'Name of the subclass of torchio.Image'
            ' that will be used to instantiate the image.'
        ),
    ),
    seed: int = typer.Option(  # noqa: B008
        None,
        '--seed',
        '-s',
        help='Seed for PyTorch random number generator.',
    ),
    verbose: bool = typer.Option(  # noqa: B008
        False,
        help='Print random transform parameters.',
    ),
    show_progress: bool = typer.Option(  # noqa: B008
        True,
        '--show-progress/--hide-progress',
        '-p/-P',
        help='Show animations indicating progress.',
    ),
):
    """Apply transform to an image.

    Example:
    $ tiotr input.nrrd RandomMotion output.nii "degrees=(-5,15) num_transforms=3" -v
    """
    # Imports are placed here so that the tool loads faster if not being run
    import torch

    import torchio.transforms as transforms
    from torchio.utils import apply_transform_to_file

    try:
        transform_class = getattr(transforms, transform_name)
    except AttributeError as error:
        message = f'Transform "{transform_name}" not found in torchio'
        raise ValueError(message) from error

    params_dict = get_params_dict_from_kwargs(kwargs)
    transform = transform_class(**params_dict)
    if seed is not None:
        torch.manual_seed(seed)
    with Progress(
        SpinnerColumn(),
        TextColumn('[progress.description]{task.description}'),
        transient=True,
        disable=not show_progress,
    ) as progress:
        progress.add_task('Applying transform', total=1)
        apply_transform_to_file(
            input_path,
            transform,
            output_path,
            verbose=verbose,
            class_=imclass,
        )


def get_params_dict_from_kwargs(kwargs):
    from torchio.utils import guess_type

    params_dict = {}
    if kwargs is not None:
        for substring in kwargs.split():
            try:
                key, value_string = substring.split('=')
            except ValueError as error:
                message = f'Arguments string "{kwargs}" not valid'
                raise ValueError(message) from error

            value = guess_type(value_string)
            params_dict[key] = value
    return params_dict


if __name__ == '__main__':
    app()



"""This is an example of a very particular case in which some modalities might
be missing for some of the subjects, as in.

Dorent et al. 2019, Hetero-Modal Variational Encoder-Decoder for Joint
Modality Completion and Segmentation
"""

import logging

import torch.nn as nn

import torchio as tio
from torchio import LabelMap
from torchio import Queue
from torchio import ScalarImage
from torchio import Subject
from torchio import SubjectsDataset
from torchio.data import UniformSampler


def main():
    # Define training and patches sampling parameters
    num_epochs = 20
    patch_size = 128
    queue_length = 100
    patches_per_volume = 5
    batch_size = 2

    # Populate a list with images
    one_subject = Subject(
        T1=ScalarImage('../BRATS2018_crop_renamed/LGG75_T1.nii.gz'),
        T2=ScalarImage('../BRATS2018_crop_renamed/LGG75_T2.nii.gz'),
        label=LabelMap('../BRATS2018_crop_renamed/LGG75_Label.nii.gz'),
    )

    # This subject doesn't have a T2 MRI!
    another_subject = Subject(
        T1=ScalarImage('../BRATS2018_crop_renamed/LGG74_T1.nii.gz'),
        label=LabelMap('../BRATS2018_crop_renamed/LGG74_Label.nii.gz'),
    )

    subjects = [
        one_subject,
        another_subject,
    ]

    subjects_dataset = SubjectsDataset(subjects)
    queue_dataset = Queue(
        subjects_dataset,
        queue_length,
        patches_per_volume,
        UniformSampler(patch_size),
    )

    # This collate_fn is needed in the case of missing modalities
    # In this case, the batch will be composed by a *list* of samples instead
    # of the typical Python dictionary that is collated by default in Pytorch
    batch_loader = tio.SubjectsLoader(
        queue_dataset,
        batch_size=batch_size,
        collate_fn=lambda x: x,
    )

    # Mock PyTorch model
    model = nn.Identity()

    for epoch_index in range(num_epochs):
        logging.info('Epoch %s', epoch_index)
        for batch in batch_loader:  # batch is a *list* here, not a dictionary
            logits = model(batch)
            logging.info([batch[idx].keys() for idx in range(batch_size)])
            logging.info(logits.shape)
    logging.info('')


if __name__ == '__main__':
    main()



from .intensity.clamp import Clamp
from .intensity.histogram_standardization import HistogramStandardization
from .intensity.mask import Mask
from .intensity.rescale import RescaleIntensity
from .intensity.z_normalization import ZNormalization
from .label.contour import Contour
from .label.keep_largest_component import KeepLargestComponent
from .label.one_hot import OneHot
from .label.remap_labels import RemapLabels
from .label.remove_labels import RemoveLabels
from .label.sequential_labels import SequentialLabels
from .spatial.copy_affine import CopyAffine
from .spatial.crop import Crop
from .spatial.crop_or_pad import CropOrPad
from .spatial.ensure_shape_multiple import EnsureShapeMultiple
from .spatial.pad import Pad
from .spatial.resample import Resample
from .spatial.resize import Resize
from .spatial.to_canonical import ToCanonical
from .spatial.to_orientation import ToOrientation

__all__ = [
    'Pad',
    'Crop',
    'Resize',
    'Resample',
    'ToCanonical',
    'ToOrientation',
    'CropOrPad',
    'CopyAffine',
    'EnsureShapeMultiple',
    'Mask',
    'RescaleIntensity',
    'Clamp',
    'ZNormalization',
    'HistogramStandardization',
    'OneHot',
    'Contour',
    'RemapLabels',
    'RemoveLabels',
    'SequentialLabels',
    'KeepLargestComponent',
]



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestToOrientation(TorchioTestCase):
    def test_invalid_orientation_length(self):
        with pytest.raises(ValueError, match='3-letter'):
            tio.ToOrientation('RA')  # Too short

    def test_invalid_orientation_characters(self):
        with pytest.raises(ValueError, match='three distinct characters'):
            tio.ToOrientation('XYZ')

    def test_missing_axis_direction(self):
        match = 'must include one character for each axis'
        with pytest.raises(ValueError, match=match):
            tio.ToOrientation('RAA')  # no S/I direction

    def test_no_change_if_already_correct(self):
        transform = tio.ToOrientation('RAS')
        subject = transform(self.sample_subject)
        self.assert_tensor_equal(subject.t1.data, self.sample_subject.t1.data)
        self.assert_tensor_equal(subject.t1.affine, self.sample_subject.t1.affine)

    def test_ras_to_las(self):
        # Step 1: Set initial orientation to RAS (default)
        ras_subject = self.sample_subject

        # Step 2: RAS -> LAS
        to_las = tio.ToOrientation('LAS')
        las_subject = to_las(ras_subject)

        self.assertEqual(las_subject.t1.orientation, ('L', 'A', 'S'))

        # Manually compute expected LAS affine
        expected_affine = ras_subject.t1.affine.copy()
        expected_affine[0, 0] = -ras_subject.t1.affine[0, 0]
        expected_affine[0, 3] = (
            ras_subject.t1.affine[0, 0] * (ras_subject.t1.spatial_shape[0] - 1)
            + ras_subject.t1.affine[0, 3]
        )

        # Check transformation validity
        flipped_data = torch.flip(ras_subject.t1.data, dims=[1])
        self.assert_tensor_almost_equal(
            las_subject.t1.data,
            flipped_data,
            check_stride=False,
        )
        self.assert_tensor_almost_equal(
            las_subject.t1.affine,
            expected_affine,
        )

    def test_ras_to_las_to_ras(self):
        # Step 1: Start with RAS orientation
        original_subject = self.sample_subject
        original_data = original_subject.t1.data.clone()
        original_affine = original_subject.t1.affine.copy()

        # Step 2: RAS -> LAS
        to_las = tio.ToOrientation('LAS')
        las_subject = to_las(original_subject)
        self.assertEqual(las_subject.t1.orientation, ('L', 'A', 'S'))

        # Step 3: LAS -> RAS
        to_ras = tio.ToOrientation('RAS')
        recovered_subject = to_ras(las_subject)
        self.assertEqual(recovered_subject.t1.orientation, ('R', 'A', 'S'))

        # Step 4: Check if data and affine are restored
        self.assert_tensor_almost_equal(
            recovered_subject.t1.data,
            original_data,
            check_stride=False,
        )
        self.assert_tensor_almost_equal(
            recovered_subject.t1.affine,
            original_affine,
        )



import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestRandomSwap(TorchioTestCase):
    def test_no_swap(self):
        transform = tio.RandomSwap(patch_size=5, num_iterations=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_swap(self):
        transform = tio.RandomSwap(patch_size=5)
        transformed = transform(self.sample_subject)
        self.assert_tensor_not_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_wrong_num_iterations_type(self):
        with pytest.raises(TypeError):
            tio.RandomSwap(num_iterations='wrong')

    def test_negative_num_iterations(self):
        with pytest.raises(ValueError):
            tio.RandomSwap(num_iterations=-1)



import numpy as np
import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestResize(TorchioTestCase):
    """Tests for `Resize`."""

    def test_one_dim(self):
        target_shape = 5
        transform = tio.Resize(target_shape)
        transformed = transform(self.sample_subject)
        for image in transformed.get_images(intensity_only=False):
            assert image.spatial_shape == 3 * (target_shape,)

    def test_all_dims(self):
        target_shape = 11, 6, 7
        transform = tio.Resize(target_shape)
        transformed = transform(self.sample_subject)
        for image in transformed.get_images(intensity_only=False):
            assert image.spatial_shape == target_shape

    def test_fix_shape(self):
        # We use values that are known to need cropping
        tensor = torch.rand(1, 8, 180, 320)
        affine = np.diag((5, 1, 1, 1))
        im = tio.ScalarImage(tensor=tensor, affine=affine)
        target = 12
        with pytest.warns(RuntimeWarning):
            result = tio.Resize(target)(im)
        assert result.spatial_shape == 3 * (target,)



import pytest

import torchio as tio

from ...utils import TorchioTestCase


class TestEnsureShapeMultiple(TorchioTestCase):
    def test_bad_method(self):
        with pytest.raises(ValueError):
            tio.EnsureShapeMultiple(1, method='bad')

    def test_pad(self):
        sample_t1 = self.sample_subject.t1
        assert sample_t1.shape == (1, 10, 20, 30)
        transform = tio.EnsureShapeMultiple(4, method='pad')
        transformed = transform(sample_t1)
        assert transformed.shape == (1, 12, 20, 32)

    def test_crop(self):
        sample_t1 = self.sample_subject.t1
        assert sample_t1.shape == (1, 10, 20, 30)
        transform = tio.EnsureShapeMultiple(4, method='crop')
        transformed = transform(sample_t1)
        assert transformed.shape == (1, 8, 20, 28)

    def test_2d(self):
        sample_t1 = self.sample_subject.t1
        sample_2d = sample_t1.data[..., :1]
        assert sample_2d.shape == (1, 10, 20, 1)
        transform = tio.EnsureShapeMultiple(4, method='crop')
        transformed = transform(sample_2d)
        assert transformed.shape == (1, 8, 20, 1)



import torchio as tio

from ...utils import TorchioTestCase


class TestRemoveLabels(TorchioTestCase):
    """Tests for `RemoveLabels`."""

    def test_remove(self):
        original_labels = (1, 2, 3, 4, 5, 6, 7)
        labels_to_remove = (1, 2, 5, 6)
        remaining_labels = (3, 4, 7)

        remove_labels = tio.RemoveLabels(labels_to_remove)

        tensor = TorchioTestCase.get_tensor_with_labels(original_labels)
        subject = tio.Subject(label=tio.LabelMap(tensor=tensor))
        transformed = remove_labels(subject)

        for removed_label in labels_to_remove:
            original_mask = subject.label.data == removed_label
            new_values = transformed.label.data[original_mask]
            self.assert_tensor_all_zeros(new_values)

        for remaining_label in remaining_labels:
            original_mask = subject.label.data == remaining_label
            original_values = subject.label.data[original_mask]
            output_values = transformed.label.data[original_mask]
            self.assert_tensor_equal(original_values, output_values)



import tempfile
from pathlib import Path

import numpy as np
import pytest
import SimpleITK as sitk
import torch

from torchio.data import ScalarImage
from torchio.data import io

from ..utils import TorchioTestCase


class TestIO(TorchioTestCase):
    """Tests for `io` module."""

    def setUp(self):
        super().setUp()
        self.nii_path = self.get_image_path('read_image')
        self.dicom_dir = self.get_tests_data_dir() / 'dicom'
        self.dicom_path = self.dicom_dir / 'IMG0001.dcm'
        string = (
            '1.5 0.18088 -0.124887 0.65072 '
            '-0.20025 0.965639 -0.165653 -11.6452 '
            '0.0906326 0.18661 0.978245 11.4002 '
            '0 0 0 1 '
        )
        tensor = torch.as_tensor(np.fromstring(string, sep=' ').reshape(4, 4))
        self.matrix = tensor

    def test_read_image(self):
        # I need to find something readable by nib but not sitk
        io.read_image(self.nii_path)

    def test_save_rgb(self):
        im = ScalarImage(tensor=torch.rand(1, 4, 5, 1))
        with pytest.warns(RuntimeWarning):
            im.save(self.dir / 'test.jpg')

    def test_read_dicom_file(self):
        tensor, _ = io.read_image(self.dicom_path)
        assert tuple(tensor.shape) == (1, 88, 128, 1)

    def test_read_dicom_dir(self):
        tensor, _ = io.read_image(self.dicom_dir)
        assert tuple(tensor.shape) == (1, 88, 128, 17)

    def test_dicom_dir_missing(self):
        with pytest.raises(FileNotFoundError):
            io._read_dicom('missing')

    def test_dicom_dir_no_files(self):
        empty = self.dir / 'empty'
        empty.mkdir()
        sitk.ProcessObject_SetGlobalWarningDisplay(False)
        with pytest.raises(FileNotFoundError):
            io._read_dicom(empty)
        sitk.ProcessObject_SetGlobalWarningDisplay(True)

    def write_read_matrix(self, suffix):
        out_path = self.dir / f'matrix{suffix}'
        io.write_matrix(self.matrix, out_path)
        matrix = io.read_matrix(out_path)
        assert torch.allclose(matrix, self.matrix)

    def test_matrix_itk(self):
        self.write_read_matrix('.tfm')
        self.write_read_matrix('.h5')

    def test_matrix_txt(self):
        self.write_read_matrix('.txt')

    def test_ensure_4d_5d(self):
        tensor = torch.rand(3, 4, 5, 1, 2)
        assert io.ensure_4d(tensor).shape == (2, 3, 4, 5)

    def test_ensure_4d_5d_t_gt_1(self):
        tensor = torch.rand(3, 4, 5, 2, 2)
        with pytest.raises(ValueError):
            io.ensure_4d(tensor)

    def test_ensure_4d_2d(self):
        tensor = torch.rand(4, 5)
        assert io.ensure_4d(tensor).shape == (1, 4, 5, 1)

    def test_ensure_4d_2d_3dims_rgb_first(self):
        tensor = torch.rand(3, 4, 5)
        assert io.ensure_4d(tensor).shape == (3, 4, 5, 1)

    def test_ensure_4d_2d_3dims_rgb_last(self):
        tensor = torch.rand(4, 5, 3)
        assert io.ensure_4d(tensor).shape == (3, 4, 5, 1)

    def test_ensure_4d_3d(self):
        tensor = torch.rand(4, 5, 6)
        assert io.ensure_4d(tensor).shape == (1, 4, 5, 6)

    def test_ensure_4d_2_spatial_dims(self):
        tensor = torch.rand(4, 5, 6)
        assert io.ensure_4d(tensor, num_spatial_dims=2).shape == (4, 5, 6, 1)

    def test_ensure_4d_3_spatial_dims(self):
        tensor = torch.rand(4, 5, 6)
        assert io.ensure_4d(tensor, num_spatial_dims=3).shape == (1, 4, 5, 6)

    def test_ensure_4d_nd_not_supported(self):
        tensor = torch.rand(1, 2, 3, 4, 5)
        with pytest.raises(ValueError):
            io.ensure_4d(tensor)

    def test_sitk_to_nib(self):
        data = np.random.rand(10, 12)
        image = sitk.GetImageFromArray(data)
        tensor, _ = io.sitk_to_nib(image)
        assert data.sum() == pytest.approx(tensor.sum())

    def test_sitk_to_affine(self):
        spacing = 1, 2, 3
        direction_lps = -1, 0, 0, 0, -1, 0, 0, 0, 1
        origin_lps = left, posterior, superior = -10, -20, 30
        image = sitk.GetImageFromArray(np.random.rand(10, 20, 30))
        image.SetDirection(direction_lps)
        image.SetSpacing(spacing)
        image.SetOrigin(origin_lps)
        origin_ras = -left, -posterior, superior
        fixture = np.diag((*spacing, 1))
        fixture[:3, 3] = origin_ras
        affine = io.get_ras_affine_from_sitk(image)
        self.assert_tensor_almost_equal(fixture, affine)


# This doesn't work as a method of the class
libs = 'sitk', 'nibabel'
parameters = []
for save_lib in libs:
    for load_lib in libs:
        for dims in 2, 3, 4:
            parameters.append((save_lib, load_lib, dims))


@pytest.mark.parametrize(('save_lib', 'load_lib', 'dims'), parameters)
def test_write_nd_with_a_read_it_with_b(save_lib, load_lib, dims):
    shape = [1, 4, 5, 6]
    if dims == 2:
        shape[-1] = 1
    elif dims == 4:
        shape[0] = 2
    tensor = torch.randn(*shape)
    affine = np.eye(4)
    tempdir = Path(tempfile.gettempdir()) / '.torchio_tests'
    tempdir.mkdir(exist_ok=True)
    path = tempdir / 'test_io.nii'
    save_function = getattr(io, f'_write_{save_lib}')
    load_function = getattr(io, f'_read_{save_lib}')
    save_function(tensor, affine, path)
    loaded_tensor, loaded_affine = load_function(path)
    TorchioTestCase.assert_tensor_equal(
        tensor.squeeze(),
        loaded_tensor.squeeze(),
        msg=f'Save lib: {save_lib}; load lib: {load_lib}; dims: {dims}',
        check_stride=False,
    )
    TorchioTestCase.assert_tensor_equal(affine, loaded_affine)


class TestNibabelToSimpleITK(TorchioTestCase):
    def setUp(self):
        super().setUp()
        self.affine = np.eye(4)

    def test_wrong_num_dims(self):
        with pytest.raises(ValueError):
            io.nib_to_sitk(np.random.rand(10, 10), self.affine)

    def test_2d_single(self):
        data = np.random.rand(1, 10, 12, 1)
        image = io.nib_to_sitk(data, self.affine)
        assert image.GetDimension() == 2
        assert image.GetSize() == (10, 12)
        assert image.GetNumberOfComponentsPerPixel() == 1

    def test_2d_multi(self):
        data = np.random.rand(5, 10, 12, 1)
        image = io.nib_to_sitk(data, self.affine)
        assert image.GetDimension() == 2
        assert image.GetSize() == (10, 12)
        assert image.GetNumberOfComponentsPerPixel() == 5

    def test_2d_3d_single(self):
        data = np.random.rand(1, 10, 12, 1)
        image = io.nib_to_sitk(data, self.affine, force_3d=True)
        assert image.GetDimension() == 3
        assert image.GetSize() == (10, 12, 1)
        assert image.GetNumberOfComponentsPerPixel() == 1

    def test_2d_3d_multi(self):
        data = np.random.rand(5, 10, 12, 1)
        image = io.nib_to_sitk(data, self.affine, force_3d=True)
        assert image.GetDimension() == 3
        assert image.GetSize() == (10, 12, 1)
        assert image.GetNumberOfComponentsPerPixel() == 5

    def test_3d_single(self):
        data = np.random.rand(1, 8, 10, 12)
        image = io.nib_to_sitk(data, self.affine)
        assert image.GetDimension() == 3
        assert image.GetSize() == (8, 10, 12)
        assert image.GetNumberOfComponentsPerPixel() == 1

    def test_3d_multi(self):
        data = np.random.rand(5, 8, 10, 12)
        image = io.nib_to_sitk(data, self.affine)
        assert image.GetDimension() == 3
        assert image.GetSize() == (8, 10, 12)
        assert image.GetNumberOfComponentsPerPixel() == 5



import torch

# Image types
INTENSITY = 'intensity'
LABEL = 'label'
SAMPLING_MAP = 'sampling_map'

# Keys for dataset samples
PATH = 'path'
TYPE = 'type'
STEM = 'stem'
DATA = 'data'
AFFINE = 'affine'
TENSOR = 'tensor'

# For aggregator
IMAGE = 'image'
LOCATION = 'location'

# For special collate function
HISTORY = 'history'

# In PyTorch convention
CHANNELS_DIMENSION = 1

# Code repository
REPO_URL = 'https://github.com/TorchIO-project/torchio/'

# Data repository
DATA_REPO = 'https://github.com/TorchIO-project/torchio-data/raw/main/data/'

# Floating point error
MIN_FLOAT_32 = torch.finfo(torch.float32).eps

# For the queue
NUM_SAMPLES = 'num_samples'



import pytest

from torchio import RandomGhosting

from ...utils import TorchioTestCase


class TestRandomGhosting(TorchioTestCase):
    """Tests for `RandomGhosting`."""

    def test_with_zero_intensity(self):
        transform = RandomGhosting(intensity=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_almost_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_zero_ghost(self):
        transform = RandomGhosting(num_ghosts=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_almost_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_ghosting(self):
        transform = RandomGhosting()
        transformed = transform(self.sample_subject)
        self.assert_tensor_not_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_intensity_range_with_negative_min(self):
        with pytest.raises(ValueError):
            RandomGhosting(intensity=(-0.5, 4))

    def test_wrong_intensity_type(self):
        with pytest.raises(ValueError):
            RandomGhosting(intensity='wrong')

    def test_negative_num_ghosts(self):
        with pytest.raises(ValueError):
            RandomGhosting(num_ghosts=-1)

    def test_num_ghosts_range_with_negative_min(self):
        with pytest.raises(ValueError):
            RandomGhosting(num_ghosts=(-1, 4))

    def test_not_integer_num_ghosts(self):
        with pytest.raises(ValueError):
            RandomGhosting(num_ghosts=(0.7, 4))

    def test_wrong_num_ghosts_type(self):
        with pytest.raises(ValueError):
            RandomGhosting(num_ghosts='wrong')

    def test_out_of_range_axis(self):
        with pytest.raises(ValueError):
            RandomGhosting(axes=3)

    def test_out_of_range_axis_in_tuple(self):
        with pytest.raises(ValueError):
            RandomGhosting(axes=(0, -1, 2))

    def test_wrong_axes_type(self):
        with pytest.raises(ValueError):
            RandomGhosting(axes=None)

    def test_out_of_range_restore(self):
        with pytest.raises(ValueError):
            RandomGhosting(restore=-1)

    def test_wrong_restore_type(self):
        with pytest.raises(ValueError):
            RandomGhosting(restore='wrong')



import pytest
import torch

import torchio as tio

from ...utils import TorchioTestCase


class TestOneHot(TorchioTestCase):
    """Tests for `OneHot`."""

    def test_one_hot(self):
        image = self.sample_subject.label
        one_hot = tio.OneHot(num_classes=3)(image)
        assert one_hot.num_channels == 3

    def test_multichannel(self):
        label_map = tio.LabelMap(tensor=torch.rand(2, 3, 3, 3) > 1)
        with pytest.raises(RuntimeError):
            tio.OneHot()(label_map)

    def test_inverse(self):
        one_hot = tio.OneHot()
        subject_one_hot = one_hot(self.sample_subject)
        subject_back = subject_one_hot.apply_inverse_transform()
        self.assert_tensor_equal(
            self.sample_subject.label.data,
            subject_back.label.data,
        )



import platform
import re
import sys

import numpy
import SimpleITK as sitk
import torch

import torchio as tio

sitk_version = re.findall('SimpleITK Version: (.*?)\n', str(sitk.Version()))[0]

print('Platform:  ', platform.platform())
print('TorchIO:   ', tio.__version__)
print('PyTorch:   ', torch.__version__)
print('SimpleITK: ', sitk_version)
print('NumPy:     ', numpy.__version__)
print('Python:    ', sys.version)



import numpy as np
import torch


class FourierTransform:
    @staticmethod
    def fourier_transform(tensor: torch.Tensor) -> torch.Tensor:
        try:
            import torch.fft

            transformed = torch.fft.fftn(tensor)
            fshift = torch.fft.fftshift(transformed)
            return fshift
        except (ModuleNotFoundError, AttributeError):
            import torch

            transformed = np.fft.fftn(tensor)
            fshift = np.fft.fftshift(transformed)
            return torch.from_numpy(fshift)

    @staticmethod
    def inv_fourier_transform(tensor: torch.Tensor) -> torch.Tensor:
        try:
            import torch.fft

            f_ishift = torch.fft.ifftshift(tensor)
            img_back = torch.fft.ifftn(f_ishift)
            return img_back
        except (ModuleNotFoundError, AttributeError):
            import torch

            f_ishift = np.fft.ifftshift(tensor)
            img_back = np.fft.ifftn(f_ishift)
            return torch.from_numpy(img_back)



import warnings

import numpy as np

from ....data.subject import Subject
from ....types import TypeSpatialShape
from ....utils import to_tuple
from ...spatial_transform import SpatialTransform
from .crop_or_pad import CropOrPad
from .resample import Resample


class Resize(SpatialTransform):
    """Resample images so the output shape matches the given target shape.

    The field of view remains the same.

    .. warning:: In most medical image applications, this transform should not
        be used as it will deform the physical object by scaling anistropically
        along the different dimensions. The solution to change an image size is
        typically applying :class:`~torchio.transforms.Resample` and
        :class:`~torchio.transforms.CropOrPad`.

    Args:
        target_shape: Tuple :math:`(W, H, D)`. If a single value :math:`N` is
            provided, then :math:`W = H = D = N`. The size of dimensions set to
            -1 will be kept.
        image_interpolation: See :ref:`Interpolation`.
        label_interpolation: See :ref:`Interpolation`.
    """

    def __init__(
        self,
        target_shape: TypeSpatialShape,
        image_interpolation: str = 'linear',
        label_interpolation: str = 'nearest',
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.target_shape = np.asarray(to_tuple(target_shape, length=3))
        self.image_interpolation = self.parse_interpolation(
            image_interpolation,
        )
        self.label_interpolation = self.parse_interpolation(
            label_interpolation,
        )
        self.args_names = [
            'target_shape',
            'image_interpolation',
            'label_interpolation',
        ]

    def apply_transform(self, subject: Subject) -> Subject:
        shape_in = np.asarray(subject.spatial_shape)
        shape_out = self.target_shape
        negative_mask = shape_out == -1
        shape_out[negative_mask] = shape_in[negative_mask]
        spacing_in = np.asarray(subject.spacing)
        spacing_out = shape_in / shape_out * spacing_in
        resample = Resample(
            spacing_out,
            image_interpolation=self.image_interpolation,
            label_interpolation=self.label_interpolation,
            **self.get_base_args(),
        )
        resampled = resample(subject)
        assert isinstance(resampled, Subject)
        # Sometimes, the output shape is one voxel too large
        # Probably because Resample uses np.ceil to compute the shape
        if not resampled.spatial_shape == tuple(shape_out):
            message = (
                f'Output shape {resampled.spatial_shape}'
                f' != target shape {tuple(shape_out)}. Fixing with CropOrPad'
            )
            warnings.warn(message, RuntimeWarning, stacklevel=2)
            crop_pad = CropOrPad(shape_out, **self.get_base_args())  # type: ignore[arg-type]
            resampled = crop_pad(resampled)
        assert isinstance(resampled, Subject)
        return resampled



import pytest

from torchio.data import PatchSampler

from ...utils import TorchioTestCase


class TestPatchSampler(TorchioTestCase):
    """Tests for `PatchSampler` class."""

    def test_bad_patch_size(self):
        with pytest.raises(ValueError):
            PatchSampler(0)
        with pytest.raises(ValueError):
            PatchSampler(-1)
        with pytest.raises(ValueError):
            PatchSampler(1.5)

    def test_extract_patch(self):
        PatchSampler(1).extract_patch(self.sample_subject, (3, 4, 5))



import os

import pytest

import torchio as tio
from torchio.datasets.medmnist import AdrenalMNIST3D
from torchio.datasets.medmnist import FractureMNIST3D
from torchio.datasets.medmnist import NoduleMNIST3D
from torchio.datasets.medmnist import OrganMNIST3D
from torchio.datasets.medmnist import SynapseMNIST3D
from torchio.datasets.medmnist import VesselMNIST3D

classes = (
    OrganMNIST3D,
    NoduleMNIST3D,
    AdrenalMNIST3D,
    FractureMNIST3D,
    VesselMNIST3D,
    SynapseMNIST3D,
)


@pytest.mark.slow
@pytest.mark.skipif('CI' in os.environ, reason='Unstable on GitHub Actions')
@pytest.mark.parametrize('class_', classes)
@pytest.mark.parametrize('split', ('train', 'val', 'test'))
def test_load_all(class_, split):
    dataset = class_(split)
    loader = tio.SubjectsLoader(
        dataset,
        batch_size=256,
    )
    for _ in loader:
        pass



import pytest

from torchio import RandomSpike

from ...utils import TorchioTestCase


class TestRandomSpike(TorchioTestCase):
    """Tests for `RandomSpike`."""

    def test_with_zero_intensity(self):
        transform = RandomSpike(intensity=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_almost_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_zero_spike(self):
        transform = RandomSpike(num_spikes=0)
        transformed = transform(self.sample_subject)
        self.assert_tensor_almost_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_with_spikes(self):
        transform = RandomSpike()
        transformed = transform(self.sample_subject)
        self.assert_tensor_not_equal(
            self.sample_subject.t1.data,
            transformed.t1.data,
        )

    def test_negative_num_spikes(self):
        with pytest.raises(ValueError):
            RandomSpike(num_spikes=-1)

    def test_num_spikes_range_with_negative_min(self):
        with pytest.raises(ValueError):
            RandomSpike(num_spikes=(-1, 4))

    def test_not_integer_num_spikes(self):
        with pytest.raises(ValueError):
            RandomSpike(num_spikes=(0.7, 4))

    def test_wrong_num_spikes_type(self):
        with pytest.raises(ValueError):
            RandomSpike(num_spikes='wrong')

    def test_wrong_intensity_type(self):
        with pytest.raises(ValueError):
            RandomSpike(intensity='wrong')



import sys

import pytest
import torch
from parameterized import parameterized

import torchio as tio
from torchio.data import UniformSampler
from torchio.utils import create_dummy_dataset

from ..utils import TorchioTestCase


class TestQueue(TorchioTestCase):
    """Tests for `queue` module."""

    def setUp(self):
        super().setUp()
        self.subjects_list = create_dummy_dataset(
            num_images=10,
            size_range=(10, 20),
            directory=self.dir,
            suffix='.nii',
            force=False,
        )

    def run_queue(self, num_workers=0, **kwargs):
        subjects_dataset = tio.SubjectsDataset(self.subjects_list)
        patch_size = 10
        sampler = UniformSampler(patch_size)
        queue_dataset = tio.Queue(
            subjects_dataset,
            max_length=6,
            samples_per_volume=2,
            sampler=sampler,
            num_workers=num_workers,
            **kwargs,
        )
        _ = str(queue_dataset)
        batch_loader = tio.SubjectsLoader(queue_dataset, batch_size=4)
        for batch in batch_loader:
            _ = batch['one_modality'][tio.DATA]
            _ = batch['segmentation'][tio.DATA]
        return queue_dataset

    def test_queue(self):
        self.run_queue(num_workers=0)

    @pytest.mark.skipif(sys.platform == 'darwin', reason='Takes too long on macOS')
    def test_queue_multiprocessing(self):
        self.run_queue(num_workers=2)

    def test_queue_no_start_background(self):
        self.run_queue(num_workers=0, start_background=False)

    @parameterized.expand([(11,), (12,)])
    def test_different_samples_per_volume(self, max_length):
        image2 = tio.ScalarImage(tensor=2 * torch.ones(1, 1, 1, 1))
        image10 = tio.ScalarImage(tensor=10 * torch.ones(1, 1, 1, 1))
        subject2 = tio.Subject(im=image2, num_samples=2)
        subject10 = tio.Subject(im=image10, num_samples=10)
        dataset = tio.SubjectsDataset([subject2, subject10])
        patch_size = 1
        sampler = UniformSampler(patch_size)
        queue_dataset = tio.Queue(
            dataset,
            max_length=max_length,
            samples_per_volume=3,  # should be ignored
            sampler=sampler,
            shuffle_patches=False,
        )
        batch_loader = tio.SubjectsLoader(queue_dataset, batch_size=6)
        tensors = [batch['im'][tio.DATA] for batch in batch_loader]
        all_numbers = torch.stack(tensors).flatten().tolist()
        assert all_numbers.count(10) == 10
        assert all_numbers.count(2) == 2

    def test_get_memory_string(self):
        queue = self.run_queue()
        memory_string = queue.get_max_memory_pretty()
        assert isinstance(memory_string, str)
